# from OCC.Core.AIS import *
# from OCC.Core.Adaptor2d import *
# from OCC.Core.Adaptor3d import *
import time

from OCC.Core.Addons import *
# from OCC.Core.AdvApp2Var import *
# from OCC.Core.AdvApprox import *
# from OCC.Core.AppBlend import *
# from OCC.Core.AppCont import *
# from OCC.Core.AppDef import *
# from OCC.Core.AppParCurves import *
# from OCC.Core.AppStd import *
# from OCC.Core.AppStdL import *
# from OCC.Core.Approx import *
# from OCC.Core.ApproxInt import *
# from OCC.Core.Aspect import *
# from OCC.Core.BOPAlgo import *
# from OCC.Core.BOPCol import *
# from OCC.Core.BOPDS import *
# from OCC.Core.BOPInt import *
# from OCC.Core.BOPTools import *
import pathlib
import tempfile
from typing import List, Optional, Union

import numpy as np
import vtk
from OCC.Core.BRep import *
from OCC.Core.BRepAdaptor import *
# from OCC.Core.BRepAlgo import *
from OCC.Core.BRepAlgoAPI import *
# from OCC.Core.BRepApprox import *
# from OCC.Core.BRepBlend import *
from OCC.Core.BRepBndLib import *
from OCC.Core.BRepBuilderAPI import *
# from OCC.Core.BRepCheck import *
# from OCC.Core.BRepClass import *
from OCC.Core.BRepClass3d import *
# from OCC.Core.BRepExtrema import *
# from OCC.Core.BRepFeat import *
# from OCC.Core.BRepFill import *
# from OCC.Core.BRepFilletAPI import *
from OCC.Core.BRepGProp import *
# from OCC.Core.BRepIntCurveSurface import *
# from OCC.Core.BRepLProp import *
# from OCC.Core.BRepLib import *
# from OCC.Core.BRepMAT2d import *
from OCC.Core.BRepMesh import *
# from OCC.Core.BRepOffset import *
from OCC.Core.BRepOffsetAPI import *
# from OCC.Core.BRepPrim import *
from OCC.Core.BRepPrimAPI import *
# from OCC.Core.BRepProj import *
from OCC.Core.BRepSweep import *
from OCC.Core.BRepTools import *
# from OCC.Core.BRepTopAdaptor import *
# from OCC.Core.BSplCLib import *
# from OCC.Core.BSplSLib import *
# from OCC.Core.BiTgte import *
# from OCC.Core.Bisector import *
# from OCC.Core.Blend import *
# from OCC.Core.BlendFunc import *
from OCC.Core.Bnd import *
# from OCC.Core.BndLib import *
# from OCC.Core.CDF import *
# from OCC.Core.CDM import *
# from OCC.Core.CPnts import *
# from OCC.Core.CSLib import *
# from OCC.Core.ChFi2d import *
# from OCC.Core.ChFi3d import *
# from OCC.Core.ChFiDS import *
# from OCC.Core.ChFiKPart import *
# from OCC.Core.Contap import *
# from OCC.Core.Convert import *
# from OCC.Core.Dico import *
# from OCC.Core.Draft import *
# from OCC.Core.DsgPrs import *
# from OCC.Core.Dynamic import *
# from OCC.Core.ElCLib import *
# from OCC.Core.ElSLib import *
# from OCC.Core.Expr import *
# from OCC.Core.ExprIntrp import *
# from OCC.Core.Extrema import *
# from OCC.Core.FEmTool import *
# from OCC.Core.FSD import *
# from OCC.Core.FairCurve import *
# from OCC.Core.FilletSurf import *
from OCC.Core.GC import *
# from OCC.Core.GCE2d import *
from OCC.Core.GCPnts import *
# from OCC.Core.GEOMAlgo import *
from OCC.Core.GProp import *
# from OCC.Core.GccAna import *
# from OCC.Core.GccEnt import *
# from OCC.Core.GccGeo import *
# from OCC.Core.GccInt import *
# from OCC.Core.GccIter import *
from OCC.Core.Geom import *
# from OCC.Core.Geom2d import *
# from OCC.Core.Geom2dAPI import *
# from OCC.Core.Geom2dAdaptor import *
# from OCC.Core.Geom2dConvert import *
# from OCC.Core.Geom2dGcc import *
# from OCC.Core.Geom2dHatch import *
# from OCC.Core.Geom2dInt import *
# from OCC.Core.Geom2dLProp import *
from OCC.Core.GeomAPI import *
# from OCC.Core.GeomAbs import *
# from OCC.Core.GeomAdaptor import *
# from OCC.Core.GeomConvert import *
# from OCC.Core.GeomFill import *
# from OCC.Core.GeomInt import *
# from OCC.Core.GeomLProp import *
# from OCC.Core.GeomLib import *
# from OCC.Core.GeomPlate import *
from OCC.Core.GeomProjLib import *
# from OCC.Core.GeomToStep import *
# from OCC.Core.GeomTools import *
# from OCC.Core.GraphDS import *
# from OCC.Core.GraphTools import *
# from OCC.Core.Graphic3d import *
# from OCC.Core.HLRAlgo import *
# from OCC.Core.HLRAppli import *
# from OCC.Core.HLRBRep import *
# from OCC.Core.HLRTopoBRep import *
# from OCC.Core.Hatch import *
# from OCC.Core.HatchGen import *
# from OCC.Core.Hermit import *
from OCC.Core.IFSelect import *
# from OCC.Core.IGESCAFControl import *
from OCC.Core.IGESControl import *
# from OCC.Core.Image import *
# from OCC.Core.IncludeLibrary import *
# from OCC.Core.IntAna import *
# from OCC.Core.IntAna2d import *
# from OCC.Core.IntCurve import *
# from OCC.Core.IntCurveSurface import *
# from OCC.Core.IntCurvesFace import *
# from OCC.Core.IntImp import *
# from OCC.Core.IntImpParGen import *
# from OCC.Core.IntPatch import *
# from OCC.Core.IntPoly import *
# from OCC.Core.IntPolyh import *
# from OCC.Core.IntRes2d import *
# from OCC.Core.IntStart import *
# from OCC.Core.IntSurf import *
# from OCC.Core.IntTools import *
# from OCC.Core.IntWalk import *
# from OCC.Core.Interface import *
# from OCC.Core.InterfaceGraphic import *
# from OCC.Core.Intf import *
# from OCC.Core.Intrv import *
# from OCC.Core.LProp import *
# from OCC.Core.LProp3d import *
# from OCC.Core.Law import *
# from OCC.Core.LocOpe import *
# from OCC.Core.LocalAnalysis import *
# from OCC.Core.MAT import *
# from OCC.Core.MAT2d import *
# from OCC.Core.MMgt import *
# from OCC.Core.Materials import *
# from OCC.Core.MeshVS import *
# from OCC.Core.Message import *
# from OCC.Core.NCollection import *
# from OCC.Core.NIS import *
# from OCC.Core.NLPlate import *
# from OCC.Core.OSD import *
# from OCC.Core.PCDM import *
# from OCC.Core.PLib import *
# from OCC.Core.Plate import *
# from OCC.Core.Plugin import *
from OCC.Core.Poly import *
from OCC.Core.Precision import *
# from OCC.Core.Primitives import *
# from OCC.Core.ProjLib import *
# from OCC.Core.Prs3d import *
# from OCC.Core.PrsMgr import *
# from OCC.Core.Quantity import *
# from OCC.Core.RWStepAP203 import *
# from OCC.Core.RWStepAP214 import *
# from OCC.Core.RWStepBasic import *
# from OCC.Core.RWStepDimTol import *
# from OCC.Core.RWStepElement import *
# from OCC.Core.RWStepFEA import *
# from OCC.Core.RWStepGeom import *
# from OCC.Core.RWStepRepr import *
# from OCC.Core.RWStepShape import *
# from OCC.Core.RWStepVisual import *
# from OCC.Core.RWStl import *
# from OCC.Core.Resource import *
# from OCC.Core.STEPCAFControl import *
# from OCC.Core.STEPConstruct import *
from OCC.Core.STEPControl import *
# from OCC.Core.STEPEdit import *
# from OCC.Core.STEPSelections import *
# from OCC.Core.Select3D import *
# from OCC.Core.SelectBasics import *
# from OCC.Core.SelectMgr import *
# from OCC.Core.ShapeAlgo import *
# from OCC.Core.ShapeAnalysis import *
# from OCC.Core.ShapeBuild import *
# from OCC.Core.ShapeConstruct import *
# from OCC.Core.ShapeCustom import *
# from OCC.Core.ShapeExtend import *
# from OCC.Core.ShapeFix import *
# from OCC.Core.ShapeProcess import *
# from OCC.Core.ShapeProcessAPI import *
# from OCC.Core.ShapeUpgrade import *
# from OCC.Core.SortTools import *
# from OCC.Core.Standard import *
# from OCC.Core.StdFail import *
# from OCC.Core.StdPrs import *
# from OCC.Core.StdSelect import *
# from OCC.Core.StepAP203 import *
# from OCC.Core.StepAP209 import *
# from OCC.Core.StepAP214 import *
# from OCC.Core.StepBasic import *
# from OCC.Core.StepDimTol import *
# from OCC.Core.StepElement import *
# from OCC.Core.StepFEA import *
# from OCC.Core.StepGeom import *
# from OCC.Core.StepRepr import *
# from OCC.Core.StepShape import *
# from OCC.Core.StepToGeom import *
# from OCC.Core.StepToTopoDS import *
# from OCC.Core.StepVisual import *
from OCC.Core.StlAPI import *
# from OCC.Core.StlTransfer import *
# from OCC.Core.Storage import *
# from OCC.Core.Sweep import *
# from OCC.Core.TColGeom import *
# from OCC.Core.TColGeom2d import *
# from OCC.Core.TColQuantity import *
from OCC.Core.TColStd import *
from OCC.Core.TColgp import *
# from OCC.Core.TCollection import *
# from OCC.Core.TDF import *
# from OCC.Core.TDataStd import *
# from OCC.Core.TDataXtd import *
# from OCC.Core.TDocStd import *
# from OCC.Core.TFunction import *
# from OCC.Core.TNaming import *
# from OCC.Core.TPrsStd import *
# from OCC.Core.TShort import *
from OCC.Core.TopAbs import *
# from OCC.Core.TopBas import *
# from OCC.Core.TopClass import *
# from OCC.Core.TopCnx import *
from OCC.Core.TopExp import *
from OCC.Core.TopLoc import *
# from OCC.Core.TopOpeBRep import *
# from OCC.Core.TopOpeBRepBuild import *
# from OCC.Core.TopOpeBRepDS import *
# from OCC.Core.TopOpeBRepTool import *
from OCC.Core.TopTools import *
# from OCC.Core.TopTrans import *
from OCC.Core.TopoDS import *
# from OCC.Core.TopoDSToStep import *
# from OCC.Core.Transfer import *
# from OCC.Core.TransferBRep import *
# from OCC.Core.UTL import *
# from OCC.Core.Units import *
# from OCC.Core.UnitsAPI import *
# from OCC.Core.V3d import *
# from OCC.Core.Visual3d import *
# from OCC.Core.Visualization import *
# from OCC.Core.Voxel import *
# from OCC.Core.XBRepMesh import *
# from OCC.Core.XCAFApp import *
# from OCC.Core.XCAFDoc import *
# from OCC.Core.XCAFPrs import *
# from OCC.Core.XSControl import *
# from OCC.Core.gce import *
from OCC.Core.gp import *
from PySide2.QtCore import *

import vmi
from numba import jit

# from OCC.Core.math import *


register_font('C:/Windows/Fonts/msyh.ttc')
register_font('C:/Windows/Fonts/simfang.ttf')
register_font('C:/Windows/Fonts/Deng.ttf')
register_font('C:/Windows/Fonts/Dengb.ttf')
register_font('C:/Windows/Fonts/Dengl.ttf')

ccType = [TopAbs_COMPOUND,
          TopAbs_COMPSOLID,
          TopAbs_SOLID,
          TopAbs_SHELL,
          TopAbs_FACE,
          TopAbs_WIRE,
          TopAbs_EDGE,
          TopAbs_VERTEX,
          TopAbs_SHAPE]

ccTypeCast = {TopAbs_COMPOUND: topods.Compound,
              TopAbs_COMPSOLID: topods.CompSolid,
              TopAbs_SOLID: topods.Solid,
              TopAbs_SHELL: topods.Shell,
              TopAbs_FACE: topods.Face,
              TopAbs_WIRE: topods.Wire,
              TopAbs_EDGE: topods.Edge,
              TopAbs_VERTEX: topods.Vertex}


def ccExplore_Recursive(sh: TopoDS_Shape, target_type: int = ccType[-1]) -> List[TopoDS_Shape]:
    """返回sh的所有下级形状，筛选目标类型target_type =

    TopAbs_COMPOUND /
    TopAbs_COMPSOLID /
    TopAbs_SOLID /
    TopAbs_SHELL /
    TopAbs_FACE /
    TopAbs_WIRE /
    TopAbs_EDGE /
    TopAbs_VERTEX /
    TopAbs_SHAPE"""
    shs, exp = [], TopExp_Explorer(sh, target_type)
    while exp.More():
        shs.append(exp.Current())
        exp.Next()
    return shs


def ccExplore_Once(sh: TopoDS_Shape) -> List[TopoDS_Shape]:
    """返回sh的直接下级形状"""
    shs, it = [], TopoDS_Iterator(sh)
    while it.More():
        shs.append(it.Value())
        it.Next()
    return shs


def ccOpenFile_IGES(file: str = None) -> Optional[TopoDS_Shape]:
    """打开IGES文件，返回形状"""
    if file is None:
        file = vmi.askOpenFile('导入 (Import) IGES', '*.igs')
        if file is None:
            return
    r = IGESControl_Reader()
    if r.ReadFile(file):
        r.TransferRoots()
        sh: TopoDS_Shape = r.OneShape()
        sh_sub = ccExplore_Once(sh)
        if sh.ShapeType() is TopAbs_COMPOUND and len(sh_sub) == 1:
            sh = sh_sub[0]
        return sh


def ccOpenFile_STEP(file: str = None) -> Optional[TopoDS_Shape]:
    """打开STEP文件，返回形状"""
    if file is None:
        file = vmi.askOpenFile('导入 (Import) STEP', '*.stp')
        if file is None:
            return
    r = STEPControl_Reader()
    if IFSelect_RetDone == r.ReadFile(file):
        r.TransferRoots()
        sh: TopoDS_Shape = r.OneShape()
        sh_sub = ccExplore_Once(sh)
        if sh.ShapeType() is TopAbs_COMPOUND and len(sh_sub) == 1:
            sh = sh_sub[0]
        return sh


def ccSaveFile_IGES(sh: TopoDS_Shape, file=None) -> bool:
    """保存形状sh为IGES文件，返回写入是否成功"""
    if file is None:
        file = vmi.askSaveFile('导出 (Export) IGES', '*.igs')
        if file is None:
            return False
    w = IGESControl_Writer()
    w.AddShape(sh)
    return w.Write(file)


def ccSaveFile_STEP(sh: TopoDS_Shape, file=None) -> bool:
    """保存形状sh为STEP文件，返回写入是否成功"""
    if file is None:
        file = vmi.askSaveFile('导出 (Export) STEP', '*.stp')
        if file is None:
            return False
    w = STEPControl_Writer()

    w.Transfer(sh, STEPControl_AsIs)
    status = w.Write(file)
    if status == IFSelect_RetDone:
        # f = pathlib.Path(file).read_bytes()
        # i = f.find((str(time.localtime().tm_mday) + 'T' + str(time.localtime().tm_hour)).encode())
        # f = f.replace(f[i - 8: i + 11], b'')
        # pathlib.Path(file).write_bytes(f)
        return True
    return False


def ccSaveFile_STL(sh: TopoDS_Shape, file=None) -> None:
    """保存形状sh为STL文件，离散化不可逆"""
    if file is None:
        file = vmi.askSaveFile('导出 (Export) STL', '*.stl')
        if file is None:
            return
    stlapi.Write(sh, file, False)


def ccPd_Sh(sh: TopoDS_Shape) -> vtk.vtkPolyData:
    """转换形状sh为网格，离散化不可逆"""
    if sh.IsNull():
        return vtk.vtkPolyData()

    pts = vtk.vtkPoints()
    polydata = vtk.vtkPolyData()
    polydata.SetPoints(pts)
    polydata.Allocate()

    # 1. 离散化
    breptools.Clean(sh)

    discrete = BRepMesh_IncrementalMesh(sh, 1e-3, True, 0.5, True)
    discrete.Perform()

    # 2. 转换Vertex
    exp = TopExp_Explorer(sh, TopAbs_VERTEX)
    while exp.More():
        vertex = topods.Vertex(exp.Current())
        exp.Next()

        pnt = BRep_Tool.Pnt(vertex)
        id = polydata.GetPoints().InsertNextPoint(pnt.X(), pnt.Y(), pnt.Z())
        polydata.InsertNextCell(vtk.VTK_VERTEX, 1, [id])

    # 3. 转换Edge
    exp = TopExp_Explorer(sh, TopAbs_EDGE)
    while exp.More():
        edge = topods.Edge(exp.Current())
        exp.Next()

        if edge.IsNull() or BRep_Tool.Degenerated(edge):
            continue

        p_on_tri = Handle_Poly_PolygonOnTriangulation()
        triangulation = Handle_Poly_Triangulation()
        loc = TopLoc_Location()
        BRep_Tool.PolygonOnTriangulation(edge, p_on_tri, triangulation, loc, 1)
        p_3d = Handle_Poly_Polygon3D()

        if p_on_tri.IsNull():
            p_3d = BRep_Tool.Polygon3D(BRep_Tool(), edge, loc)
            p_3d = Handle_Poly_Polygon3D() if p_3d is None else p_3d

        if p_3d.IsNull() and p_on_tri.IsNull():
            continue

        trsf = loc.Transformation()

        if not p_3d.IsNull():
            n = p_3d.NbNodes()
            pnts = p_3d.Nodes()
            ids = list(range(1, n + 1))
        else:
            n = p_on_tri.NbNodes()
            pnts = triangulation.Nodes()
            ids = p_on_tri.Nodes()
            ids = [ids.Value(i) for i in range(ids.Lower(), ids.Upper() + 1)]

        if n > 1:
            idList = vtk.vtkIdList()
            for i in ids:
                pnt = pnts.Value(i)
                pnt.Transform(trsf)
                id = polydata.GetPoints().InsertNextPoint(pnt.X(), pnt.Y(), pnt.Z())
                idList.InsertNextId(id)
            polydata.InsertNextCell(vtk.VTK_POLY_LINE, idList)

    # 4. 转换Face并合并
    exp = TopExp_Explorer(sh, TopAbs_FACE)
    if exp.More():
        with tempfile.TemporaryDirectory() as p:
            p = pathlib.Path(p) / '.stl'
            stlapi.Write(sh, str(p), False)

            r = vtk.vtkSTLReader()
            r.SetFileName(str(p))
            r.Update()

            a = vtk.vtkAppendPolyData()
            a.AddInputData(r.GetOutput())
            a.AddInputData(polydata)
            a.Update()

            c = vtk.vtkCleanPolyData()
            c.SetInputData(a.GetOutput())
            c.Update()

            polydata.DeepCopy(a.GetOutput())

    return polydata


def ccSh_Pd(pd: vtk.vtkPolyData) -> TopoDS_Shape:
    """转换网格pd为形状，点单元转换为TopoDS_Vertex，线单元转换为TopoDS_Line，多边形单元转换为TopoDS_Face"""
    builder, compound = BRep_Builder(), TopoDS_Compound()
    builder.MakeCompound(compound)

    triangles = vtk.vtkTriangleFilter()
    triangles.SetInputData(pd)
    triangles.Update()
    triangles = triangles.GetOutput()

    for i in range(triangles.GetNumberOfCells()):
        cell: vtk.vtkCell = triangles.GetCell(i)
        if cell.GetCellType() == vtk.VTK_TRIANGLE:
            ids: vtk.vtkIdList = cell.GetPointIds()
            pts = [triangles.GetPoint(ids.GetId(0)),
                   triangles.GetPoint(ids.GetId(1)),
                   triangles.GetPoint(ids.GetId(2))]
            face = ccFace(ccWire(ccSegments(pts, True)))
            builder.Add(compound, face)

    sewing = BRepBuilderAPI_Sewing()
    sewing.Load(compound)
    sewing.Perform()

    sh = sewing.SewedShape()

    if sh.ShapeType() is TopAbs_SHELL and sh.Closed():
        return ccSolid(sh)
    else:
        return sh


def ccPnt(arg: Union[np.ndarray, List[float]]) -> gp_Pnt:
    """构造几何点"""
    return gp_Pnt(arg[0], arg[1], arg[2])


def ccVec(arg: Union[np.ndarray, List[float]]) -> gp_Vec:
    """构造几何向量"""
    return gp_Vec(arg[0], arg[1], arg[2])


def ccDir(arg: Union[np.ndarray, List[float]]) -> gp_Dir:
    """构造几何方向"""
    return gp_Dir(arg[0], arg[1], arg[2])


def ccAx2(origin: Union[np.ndarray, List[float]], z: Union[np.ndarray, List[float]],
          x: Union[np.ndarray, List[float]] = None) -> gp_Ax2:
    """
    构造几何坐标系

    :param origin: 原点
    :param z: Z轴向量
    :param x: X轴向量
    :return: 几何坐标系
    """
    args = [ccPnt(origin), ccDir(z)]
    if x is not None:
        args.append(ccDir(x))
    return gp_Ax2(*args)


def ccSegment(pt0: Union[np.ndarray, List[float]], pt1: Union[np.ndarray, List[float]]) -> Geom_TrimmedCurve:
    """
    构造线段

    :param pt0: 起点
    :param pt1: 终点
    :return: 线段
    """
    pt0 = pt0 if isinstance(pt0, gp_Pnt) else ccPnt(pt0)
    pt1 = pt1 if isinstance(pt1, gp_Pnt) else ccPnt(pt1)
    return GC_MakeSegment(pt0, pt1).Value()


def ccSegments(pts: List[Union[np.ndarray, List[float]]], closed: bool = False) -> List[Geom_TrimmedCurve]:
    """
    构造线段列表

    :param pts: 点列
    :param closed: 是否闭合
    :return: 线段列表
    """
    pnts_ = [pnt for pnt in pts]
    if closed:
        pnts_.append(pts[0])
    return [ccSegment(pnts_[i], pnts_[i + 1]) for i in range(len(pnts_) - 1)]


def ccBSpline(pts: List[Union[np.ndarray, List[float]]], tgs: List[Union[np.ndarray, List[float]]] = None,
              tgflags: List[bool] = None, closed: bool = False, scale: bool = True,
              tol: float = precision.Confusion()) -> Optional[Geom_BSplineCurve]:
    """
    构造B样条

    :param pts: 点列
    :param tgs: 切向量列表
    :param tgflags: 切向量有效状态列表
    :param closed: 是否闭合
    :param scale: 是否自动调节尺度
    :param tol: 误差限
    :return: B样条
    """
    if len(pts) < 2:
        return None

    if pts is not None and not isinstance(pts, TColgp_HArray1OfPnt):
        _ = TColgp_HArray1OfPnt(1, len(pts))
        for i in range(len(pts)):
            if isinstance(pts[i], gp_Pnt):
                _.SetValue(i + 1, pts[i])
            else:
                _.SetValue(i + 1, ccPnt(pts[i]))
        pts = _

    if tgs is not None and not isinstance(tgs, TColgp_Array1OfVec):
        _ = TColgp_Array1OfVec(1, len(tgs))
        for i in range(len(tgs)):
            if isinstance(pts[i], gp_Vec):
                _.SetValue(i + 1, tgs[i])
            else:
                _.SetValue(i + 1, ccVec(tgs[i]))
        tgs = _

    if tgs is not None and tgflags is None:
        tgflags = [True for _ in tgs]

    if tgflags is not None and not isinstance(tgflags, TColStd_HArray1OfBoolean):
        _ = TColStd_HArray1OfBoolean(1, len(tgflags))
        for i in range(len(tgflags)):
            _.SetValue(i + 1, tgflags[i])
        tgflags = _

    lower, upper = pts.Lower(), pts.Upper() + (1 if closed else 0)
    knots = TColStd_HArray1OfReal(lower, upper)
    knots.SetValue(lower, 0)

    for i in range(lower, upper):
        if closed and i == upper - 1:
            var = pts.Value(i).Distance(pts.Value(lower)) ** 0.5
        else:
            var = pts.Value(i).Distance(pts.Value(i + 1)) ** 0.5
        knots.SetValue(i + 1, knots.Value(i) + var)

    curve = GeomAPI_Interpolate(pts, knots, closed, tol)
    if tgs is not None:
        curve.Load(tgs, tgflags, scale)
    curve.Perform()
    return curve.Curve() if curve.IsDone() else None


def ccBSpline_Kochanek(pts: List[Union[np.ndarray, List[float]]], closed: bool = False,
                       tol: float = precision.Confusion()) -> Optional[Geom_BSplineCurve]:
    if len(pts) < 2:
        return None

    points = vtk.vtkPoints()
    for pt in pts:
        points.InsertNextPoint(pt)

    ks = [vtk.vtkKochanekSpline(), vtk.vtkKochanekSpline(), vtk.vtkKochanekSpline()]

    ps = vtk.vtkParametricSpline()
    ps.SetXSpline(ks[0])
    ps.SetYSpline(ks[1])
    ps.SetZSpline(ks[2])
    ps.SetPoints(points)
    ps.SetClosed(1 if closed else 0)
    ps.SetParameterizeByLength(1)
    ps.SetLeftConstraint(1)
    ps.SetRightConstraint(1)

    n = 100 * len(pts)

    pf = vtk.vtkParametricFunctionSource()
    pf.SetParametricFunction(ps)
    pf.SetUResolution(n)
    pf.SetVResolution(n)
    pf.SetWResolution(n)
    pf.Update()

    pts_, points = [], pf.GetOutput().GetPoints()

    for i in range(points.GetNumberOfPoints() - (1 if closed else 0)):
        pts_.append(points.GetPoint(i))
    pts_[0] = pts[0].copy()
    if not closed:
        pts_[-1] = pts[-1].copy()

    curve = ccBSpline(pts_, closed=closed, tol=tol)

    obb = vmi.pdOBB_Pts(points)
    if 0 < obb['size'][2] < tol:
        curve = ccCurveOnPlane(curve, obb['center'], obb['axis'][2])
    return curve


def ccCurveOnPlane(curve, origin: Union[np.ndarray, List[float]], normal: Union[np.ndarray, List[float]],
                   vt: Union[np.ndarray, List[float]] = None) -> Geom_Curve:
    if vt is None:
        vt = normal
    return geomprojlib.ProjectOnPlane(curve, ccPlane(origin, normal), ccDir(vt), True)


def ccCircle(radius: float, origin: Union[np.ndarray, List[float]],
             z: Union[np.ndarray, List[float]], x: Union[np.ndarray, List[float]] = None) -> Optional[TopoDS_Shape]:
    origin = [origin[0], origin[1], origin[2]]
    z = [z[0], z[1], z[2]]
    if x is not None:
        x = [x[0], x[1], x[2]]
    mk = GC_MakeCircle(ccAx2(origin, z, x), radius)
    return mk.Value() if mk.IsDone() else None


def ccRect(origin: Union[np.ndarray, List[float]], x: Union[np.ndarray, List[float]], y: Union[np.ndarray, List[float]],
           xl: float, yl: float) -> List[Geom_Curve]:
    origin, x, y = np.array(origin), np.array(x), np.array(y)
    pts = [origin,
           origin + xl * x,
           origin + xl * x, + yl * y,
           origin + yl * y]
    return ccSegments(pts)


def ccPolyline_Regular(radius: float, origin: Union[np.ndarray, List[float]], z: Union[np.ndarray, List[float]],
                       x: Union[np.ndarray, List[float]], side_num=120) -> List[Geom_Curve]:
    pts, cs = [], vmi.CS4x4(axis0=x, axis2=z, origin=origin)
    cs.orthogonalize([1])

    for i in range(side_num):
        cs = cs.rotate(360 / side_num, cs.axis(2))
        pt = cs.mpt([radius, 0, 0])
        pts.append(pt)
    return ccSegments(pts, True)


def ccPlane(origin, normal):
    mk = GC_MakePlane(ccPnt(origin), ccDir(normal))
    return mk.Value() if mk.IsDone() else None


def ccVertex(pt):
    pt = pt if isinstance(pt, gp_Pnt) else ccPnt(pt)
    mk = BRepBuilderAPI_MakeVertex(pt)
    mk.Build()
    return mk.Vertex() if mk.IsDone() else None


def ccEdge(curve):
    mk = BRepBuilderAPI_MakeEdge(curve)
    mk.Build()
    return mk.Edge() if mk.IsDone() else None


def ccEdges(curves):
    return [ccEdge(curve) for curve in curves]


def ccWire(edges):
    edges_ = edges
    if not hasattr(edges_, '__iter__'):
        edges_ = [edges_]
    edges_ = [edge if isinstance(edge, TopoDS_Edge) else ccEdge(edge) for edge in edges_]
    mk = BRepBuilderAPI_MakeWire()
    for edge in edges_:
        mk.Add(edge)
    mk.Build()
    return mk.Wire() if mk.IsDone() else None


def ccWires(edges):
    return [ccWire(edge) for edge in edges]


def ccFace(wire):
    mk = BRepBuilderAPI_MakeFace(wire)
    mk.Build()
    return mk.Face() if mk.IsDone() else None


def ccFaceOnPlane(wire, origin, normal):
    mk = BRepBuilderAPI_MakeFace(gp_Pln(ccPnt(origin), ccDir(normal)), wire)
    mk.Build()
    return mk.Face() if mk.IsDone() else None


def ccLoft(sections: List[TopoDS_Wire], ruled=False, redirect=False, solid=False):
    mk = BRepOffsetAPI_ThruSections(solid, ruled)
    mk.CheckCompatibility(redirect)
    for section in sections:
        if isinstance(section, TopoDS_Edge):
            section = ccWire(section)
        if isinstance(section, TopoDS_Wire):
            mk.AddWire(section)
        if isinstance(section, TopoDS_Vertex):
            mk.AddVertex(section)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccSew(shs: List[TopoDS_Shape]):
    mk = BRepBuilderAPI_Sewing()
    for sh in shs:
        mk.Add(sh)
    mk.Perform()
    return mk.SewedShape()


def ccSolid(shells):
    if isinstance(shells, TopoDS_Shell):
        shells = [shells]
    mk = BRepBuilderAPI_MakeSolid()
    for shell in shells:
        if shell.ShapeType() is TopAbs_SHELL:
            mk.Add(topods.Shell(shell))
    if mk.IsDone():
        s = mk.Solid()

        if ccInside_Pt(s):
            s.Reverse()
        return s


def ccBoolean_Union(sh: List[TopoDS_Shape]) -> Optional[TopoDS_Shape]:
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Fuse()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Union', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccBoolean_Difference(sh: List[TopoDS_Shape]) -> Optional[TopoDS_Shape]:
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Cut()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Difference', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccBoolean_Intersection(sh: List[TopoDS_Shape]) -> Optional[TopoDS_Shape]:
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Common()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccBoolean_Intersection', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccSection(sh: List[TopoDS_Shape]) -> Optional[TopoDS_Shape]:
    if len(sh) == 0:
        return None
    elif len(sh) == 1:
        return sh[0]
    elif len(sh) > 1:
        sh = [sh[:1], sh[1:]]
        shape_list = [TopTools_ListOfShape(), TopTools_ListOfShape()]

        for j in range(2):
            for i in sh[j]:
                shape_list[j].Append(i)

        mk = BRepAlgoAPI_Section()
        mk.SetArguments(shape_list[0])
        mk.SetTools(shape_list[1])
        mk.SetRunParallel(True)
        mk.Build()

        if not mk.IsDone() or mk.ErrorStatus() or mk.WarningStatus():
            raise Exception('ccSection', mk.ErrorStatus(), mk.WarningStatus())
        return mk.Shape()


def ccCylinder(radius: float, height: float, origin: Union[np.ndarray, List[float]],
               z: Union[np.ndarray, List[float]], x: Union[np.ndarray, List[float]] = None) -> Optional[TopoDS_Shape]:
    origin = [origin[0], origin[1], origin[2]]
    z = [z[0], z[1], z[2]]
    if x is not None:
        x = [x[0], x[1], x[2]]
    mk = BRepPrimAPI_MakeCylinder(ccAx2(origin, z, x), radius, height)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccSphere(radius: float, origin: Union[np.ndarray, List[float]],
             z: Union[np.ndarray, List[float]], x: Union[np.ndarray, List[float]] = None) -> Optional[TopoDS_Shape]:
    origin = [origin[0], origin[1], origin[2]]
    z = [z[0], z[1], z[2]]
    if x is not None:
        x = [x[0], x[1], x[2]]
    mk = BRepPrimAPI_MakeSphere(ccAx2(origin, z, x), radius)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccSphere_Half(radius: float, origin: Union[np.ndarray, List[float]],
                  z: Union[np.ndarray, List[float]], x: Union[np.ndarray, List[float]] = None) -> Optional[
    TopoDS_Shape]:
    origin = [origin[0], origin[1], origin[2]]
    z = [z[0], z[1], z[2]]
    if x is not None:
        x = [x[0], x[1], x[2]]
    mk = BRepPrimAPI_MakeSphere(ccAx2(origin, z, x), radius, np.pi)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccPrism(sh, vt, l):
    vt = np.array(vt)
    vec = l * vt / np.linalg.norm(vt)
    vec = gp_Vec(vec[0], vec[1], vec[2])
    mk = BRepPrimAPI_MakePrism(sh, vec)
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccRevolve(sh, pt, vt, angle=360):
    pt = np.array(pt)
    vt = np.array(vt)
    mk = BRepPrimAPI_MakeRevol(sh, gp_Ax1(ccPnt(pt), ccDir(vt)), np.deg2rad(angle), True)
    return mk.Shape() if mk.IsDone() else None


def ccOffset(sh, offset: float):
    mk = BRepOffsetAPI_MakeOffsetShape(sh, offset, precision.Confusion())
    mk.Build()
    return mk.Shape() if mk.IsDone() else None


def ccBounds(shs: Union[TopoDS_Shape, List[TopoDS_Shape]]):
    if isinstance(shs, TopoDS_Shape):
        shs = [shs]
    b = Bnd_Box()
    for sh in shs:
        brepbndlib.Add(sh, b)
    return np.array(b.Get())


def ccCenter(shs: Union[TopoDS_Shape, List[TopoDS_Shape]]):
    if isinstance(shs, TopoDS_Shape):
        shs = [shs]
    b = ccBounds(shs)
    return 0.5 * np.array([b[0] + b[1], b[2] + b[3], b[4] + b[5]])


def ccDiagonal(shs: Union[TopoDS_Shape, List[TopoDS_Shape]]):
    if isinstance(shs, TopoDS_Shape):
        shs = [shs]
    b = ccBounds(shs)
    return np.linalg.norm(np.array([b[1] - b[0], b[3] - b[2], b[5] - b[4]]))


def ccRayCast(sh, pt, vt):
    """把pt沿vt方向投射到sh上"""
    pt, vt = np.array(pt), np.array(vt)

    def dmax(s0, s1):
        b = [Bnd_Box(), Bnd_Box()]
        brepbndlib.Add(s0, b[0])
        brepbndlib.Add(s1, b[1])

        min = [[b[0].Get()[i] for i in range(3)],
               [b[1].Get()[i] for i in range(3)]]
        max = [[b[0].Get()[i] for i in range(3, 6)],
               [b[1].Get()[i] for i in range(3, 6)]]
        xyz = [[gp_XYZ(min[0][0], min[0][1], min[0][2]),
                gp_XYZ(max[0][0], max[0][1], max[0][2])],
               [gp_XYZ(min[1][0], min[1][1], min[1][2]),
                gp_XYZ(max[1][0], max[1][1], max[1][2])]]
        delta = [xyz[0][1].Subtracted(xyz[0][0]),
                 xyz[1][1].Subtracted(xyz[1][0])]

        def dmin(s0, s1):
            b = [Bnd_Box(), Bnd_Box()]
            brepbndlib.Add(s0, b[0])
            brepbndlib.Add(s1, b[1])
            return b[0].Distance(b[1])

        return delta[0].Modulus() + delta[1].Modulus() + dmin(s0, s1)

    d = dmax(ccVertex(pt), sh)
    vtmax = d * vt / np.linalg.norm(vt)

    sweep = BRepSweep_Prism(ccVertex(pt), vtmax)
    section = ccSection([sh, sweep.Shape()])

    origin = ccPnt(pt)

    pts = {}
    for s in ccExplore_Recursive(section, TopAbs_VERTEX):
        pnt = BRep_Tool.Pnt(s)
        pts[pnt.Distance(origin)] = [pnt.X(), pnt.Y(), pnt.Z()]
    return [pts[i] for i in sorted(pts)]


def ccUniformPts_Edge(edge: TopoDS_Shape, n: int):
    n = int(n)
    if n < 2:
        return
    curve = BRepAdaptor_Curve(edge)
    mk = GCPnts_QuasiUniformAbscissa(curve, n)
    if mk.IsDone():
        pts = [curve.Value(mk.Parameter(i)) for i in range(1, n + 1)]
        return [np.array([pt.X(), pt.Y(), pt.Z()]) for pt in pts]


def ccEdge_FirstLastVertex(edge):
    pts = [TopoDS_Vertex(), TopoDS_Vertex()]
    topexp.Vertices(edge, pts[0], pts[1])
    return pts


def ccEdge_FirstLastPoint(edge):
    pts = ccEdge_FirstLastVertex(edge)
    pts = [BRep_Tool.Pnt(pt) for pt in pts]
    pts = [[pt.X(), pt.Y(), pt.Z()] for pt in pts]
    return pts


def ccLength(sh):
    p = GProp_GProps()
    brepgprop.LinearProperties(sh, p)
    return p.Mass()


def ccArea(sh):
    p = GProp_GProps()
    brepgprop.SurfaceProperties(sh, p)
    return p.Mass()


def ccInside_Pt(sh, pt=None):
    c = BRepClass3d_SolidClassifier(sh)
    if pt is None:
        c.PerformInfinitePoint(precision.Confusion())
    else:
        c.Perform(ccPnt(pt), precision.Confusion())

    if c.State() == TopAbs_IN or c.State() == TopAbs_ON:
        return True
    return False


def ccMatchSolid_Thick(closed_pts: List[Union[np.ndarray, List[float]]], match_vt: Union[np.ndarray, List[float]],
                       target_poly: vtk.vtkPolyData, thick: Optional[float] = None, safe_dist: float = 2,
                       return_bottom_pd: Optional[List] = None) -> TopoDS_Shape:
    """
    匹配closed_pts包围的区域到面网格模型，沿match_vt方向拉伸thick距离创建匹配实体

    :param closed_pts: 首尾相连包围区域的点列
    :param match_vt: 匹配的投射方向
    :param target_poly: 匹配的模型
    :param thick: 匹配实体的厚度，thick=None时上表面为平面
    :param safe_dist: 匹配时忽略的最大孔洞半径
    :param return_bottom_pd: 返回匹配面vtkPolyData
    :return: 与模型相匹配的实体
    """
    cs = vmi.CS4x4_Vt(2, match_vt)
    z_pt = {cs.inv().mpt(pt)[2]: pt for pt in closed_pts}
    zmin = min(z_pt)

    top_pts = [vmi.ptOnPlane(pt, z_pt[zmin], match_vt) for pt in closed_pts]
    top_pd = vmi.pdTriangulate_Polyline(top_pts)
    top_pd = vmi.pdRemesh(top_pd)

    bottom_pd = vmi.pdApproximate_Pd(top_pd, match_vt, target_poly, safe_dist)
    # bottom_pd = vmi.pdSmooth_Laplacian(bottom_pd)

    if thick is not None:
        t = vtk.vtkTransform()
        t.Translate(-thick * match_vt)
        top_pd = vmi.pdTransform(bottom_pd, t)

    top_pts = vmi.pdBorder_Pts(top_pd)
    top_wire = ccWire(ccSegments(top_pts, True))
    top_face = ccSh_Pd(top_pd)

    bottom_pts = vmi.pdBorder_Pts(bottom_pd)
    bottom_wire = ccWire(ccSegments(bottom_pts, True))
    bottom_face = ccSh_Pd(bottom_pd)

    profile = ccLoft([top_wire, bottom_wire], True, True)
    sh = ccSolid(ccSew([top_face, profile, bottom_face]))

    if return_bottom_pd is not None:
        return_bottom_pd.clear()
        return_bottom_pd.append(bottom_pd)
    return sh


def ccMatchSolid_Blocks(closed_pts: List[Union[np.ndarray, List[float]]], match_vt: Union[np.ndarray, List[float]],
                        target_poly: vtk.vtkPolyData, block_size: float, safe_dist: float = 1,
                        return_bottom_pd: Optional[List] = None) -> TopoDS_Shape:
    """
    匹配closed_pts包围的区域到面网格模型，沿match_vt方向投射一群三棱柱状小块组合创建匹配实体

    :param closed_pts: 首尾相连包围区域的点列
    :param match_vt: 匹配的投射方向
    :param target_poly: 匹配的模型
    :param block_size: 柱状小块的尺寸
    :param safe_dist: 匹配时忽略的最大孔洞半径
    :param return_bottom_pd: 返回匹配面vtkPolyData
    :return: 与模型相匹配的实体
    """
    cs = vmi.CS4x4_Vt(2, match_vt)
    z_pt = {cs.inv().mpt(pt)[2]: pt for pt in closed_pts}
    zmin = min(z_pt)

    top_pts = [vmi.ptOnPlane(pt, z_pt[zmin], match_vt) for pt in closed_pts]
    top_pd = vmi.pdTriangulate_Polyline(top_pts)
    top_pd = vmi.pdRemesh(top_pd, block_size)

    bottom_pd = vmi.pdApproximate_Pd(top_pd, match_vt, target_poly, safe_dist)
    bottom_pd = vmi.pdTriangle(bottom_pd, 0, 0)
    if return_bottom_pd is not None:
        return_bottom_pd.clear()
        return_bottom_pd.append(bottom_pd)

    cs = vmi.CS4x4_Vt(0, match_vt)

    histogram = []
    for i in range(bottom_pd.GetNumberOfCells()):
        if bottom_pd.GetCell(i).GetCellType() != vtk.VTK_TRIANGLE:
            continue

        ids = [bottom_pd.GetCell(i).GetPointId(_) for _ in range(3)]
        bottom_pts = [np.array(bottom_pd.GetPoint(_)) for _ in ids]
        bottom_wire = ccWire(ccSegments(bottom_pts, True))
        bottom_face = ccFace(bottom_wire)

        top_origin = bottom_pts[0]
        for pt in bottom_pts[1:]:
            if np.dot(match_vt, top_origin - pt) > 0:
                top_origin = pt
        top_origin -= block_size * match_vt
        top_origin = cs.inv().mpt(top_origin)
        top_origin[0] = np.floor(top_origin[0] / block_size) * block_size
        top_origin = cs.mpt(top_origin)

        top_pts = [vmi.ptOnPlane(pt, top_origin, match_vt) for pt in bottom_pts]
        top_wire = ccWire(ccSegments(top_pts, True))
        top_face = ccFace(top_wire)

        profile = ccLoft([top_wire, bottom_wire], True)
        sh = ccSolid(ccSew([top_face, profile, bottom_face]))
        histogram.append(sh)

    return ccBoolean_Union(histogram)


def ccMatchSolid_Prism(closed_pts: List[Union[np.ndarray, List[float]]], match_vt: Union[np.ndarray, List[float]],
                       match_dmax: float, target_poly: vtk.vtkPolyData,
                       return_bottom_pd: Optional[List] = None) -> TopoDS_Shape:
    """
    匹配closed_pts包围的区域到面网格模型，得到匹配实体

    :param closed_pts: 首尾相连包围区域的点列
    :param match_vt: 匹配的投射方向
    :param target_poly: 匹配的模型
    :param match_dmax: vt方向最后方点的最大投射距离
    :param thick_vt: 匹配实体的厚度拉伸向量
    :return: 与模型相匹配的实体
    """

    cs = vmi.CS4x4_Vt(2, match_vt)
    z_pt = {cs.inv().mpt(pt)[2]: pt for pt in closed_pts}
    zmin = min(z_pt)

    top_pts = [vmi.ptOnPlane(pt, z_pt[zmin], match_vt) for pt in closed_pts]
    top_pd = vmi.pdTriangulate_Polyline(top_pts)
    top_pd = vmi.pdRemesh(top_pd)

    bottom_pd = vmi.pdRayCast_Pd(top_pd, match_vt, target_poly, match_dmax)
    bottom_pd = vmi.pdRemesh(bottom_pd)

    bottom_pts = vmi.pdBorder_Pts(bottom_pd)
    bottom_wire = ccWire(ccSegments(bottom_pts, True))
    bottom_face = ccSh_Pd(bottom_pd)

    top_pts = [vmi.ptOnPlane(pt, z_pt[zmin], match_vt) for pt in bottom_pts]
    top_wire = ccWire(ccSegments(top_pts, True))
    top_face = ccFaceOnPlane(top_wire, z_pt[zmin], match_vt)
    top_wire = ccExplore_Once(top_face)[0]

    profile = ccLoft([top_wire, bottom_wire], True, True)
    shape = ccSolid(ccSew([top_face, profile, bottom_face]))

    if return_bottom_pd is not None:
        return_bottom_pd.clear()
        return_bottom_pd.append(bottom_pd)
    return shape


def ccMatchSolid_Rect(target_poly: vtk.vtkPolyData, cs: vmi.CS4x4, nx: int, ny: int, res: float,
                      z_min: float, z_max: float, smooth: float = 0,
                      return_mold_inv: Optional[List] = None, return_z_array: Optional[List] = None):
    if not nx >= 1:
        raise Exception('nx={}'.format(nx))
    if not ny >= 1:
        raise Exception('ny={}'.format(ny))
    if not res > 0:
        raise Exception('res={}'.format(res))
    if not z_min > 0:
        raise Exception('thick_min={}'.format(z_min))
    if not z_max > 0:
        raise Exception('ray_max={}'.format(z_max))

    target_kd = vtk.vtkKdTreePointLocator()
    target_kd.SetDataSet(target_poly)
    target_kd.BuildLocator()

    target_obb = vtk.vtkOBBTree()
    target_obb.SetDataSet(target_poly)
    target_obb.BuildLocator()

    import time
    time_start = time.time()

    mold_pts_list, mold_inv_pts_list = [], []
    zs = np.zeros((2 * ny + 1, 2 * nx + 1))
    for j in range(2 * ny + 1):
        mold_pts, mold_inv_pts = [], []
        for i in range(2 * nx + 1):
            pt = cs.mpt([res * (i - nx), res * (j - ny), 0])
            mold_pts.append(pt.copy())
            mold_inv_pts.append(pt.copy())

            d = vmi.pdRayDistance_Pt(pt, cs.axis(2), target_poly, target_kd, target_obb)
            zs[j][i] = z_max if d is None else d
            zs[j][i] = min(max(zs[j][i], z_min), z_max)

        mold_pts.append(cs.mpt([res * nx, res * (j - ny), z_min]))
        mold_pts.append(cs.mpt([res * -nx, res * (j - ny), z_min]))
        mold_pts_list.append(mold_pts)

        mold_inv_pts.append(cs.mpt([res * nx, res * (j - ny), z_max + z_min]))
        mold_inv_pts.append(cs.mpt([res * -nx, res * (j - ny), z_max + z_min]))
        mold_inv_pts_list.append(mold_inv_pts)

    @jit(nopython=True)
    def smooth_zs(smooth, zs):
        for k in range(int(1e4)):
            delta_max = 0
            zs_ = zs.copy()
            for j in range(0, 2 * ny + 1):
                for i in range(0, 2 * nx + 1):
                    jmin, jmax = min(max(j - 1, 0), 2 * ny), min(max(j + 1, 0), 2 * ny)
                    imin, imax = min(max(i - 1, 0), 2 * nx), min(max(i + 1, 0), 2 * nx)
                    neighbors = zs_[jmin:(jmax + 1), imin:(imax + 1)]

                    if zs_[j][i] - neighbors.mean() > smooth:
                        zs[j][i] = neighbors.mean()
                    delta_max = max(delta_max, zs_[j][i] - neighbors.mean())
            if delta_max < smooth:
                break

    if smooth > 0:
        smooth_zs(smooth, zs)

    # mold inverse
    if return_mold_inv is not None:
        for j in range(0, 2 * ny + 1):
            for i in range(0, 2 * nx + 1):
                mold_inv_pts_list[j][i] += zs[j][i] * cs.axis(2)

        wires = [ccWire(ccSegments(x_pts, True)) for x_pts in mold_inv_pts_list]
        # wires = []
        # for x_pts in mold_inv_pts_list:
        #     spline = ccEdge(ccBSpline(x_pts[:-2]))
        #     lines = ccEdges(ccSegments([*x_pts[-3:], x_pts[0]]))
        #     wires.append(ccWire([spline, *lines]))

        mold_inv = ccLoft(wires, ruled=True, solid=True)
        return_mold_inv.clear()
        return_mold_inv.append(mold_inv)

    if return_z_array is not None:
        return_z_array.clear()
        return_z_array.append(zs)

    # mold
    for j in range(0, 2 * ny + 1):
        for i in range(0, 2 * nx + 1):
            mold_pts_list[j][i] += zs[j][i] * cs.axis(2)

    wires = [ccWire(ccSegments(x_pts, True)) for x_pts in mold_pts_list]
    # wires = []
    # for x_pts in mold_pts_list:
    #     spline = ccEdge(ccBSpline(x_pts[:-2]))
    #     lines = ccEdges(ccSegments([*x_pts[-3:], x_pts[0]]))
    #     wires.append(ccWire([spline, *lines]))
    mold = ccLoft(wires, ruled=True, solid=True)
    return mold


def ccTransform_Translation(sh: TopoDS_Shape, translation: Union[np.ndarray, List[float]]) -> TopoDS_Shape:
    trsf = gp_Trsf()
    trsf.SetTranslation(ccVec(translation))
    return BRepBuilderAPI_Transform(sh, trsf, True).Shape()



import sys
from threading import Thread

import numpy as np
from PySide2.QtCore import *

import vmi

W = vmi.View()
vmi.appwindow(W)


def NoButtonEnter(**kwargs):
    if kwargs['picked'] in Actor.values():
        kw = [kw for kw in Actor if Actor[kw] is kwargs['picked']][0]

        if kw in Actor:
            Text['显示名称'].draw_text('{}'.format(kw))
        return Actor[kw].mouseEnter(**kwargs)


def NoButtonLeave(**kwargs):
    if kwargs['picked'] in Actor.values():
        kw = [kw for kw in Actor if Actor[kw] is kwargs['picked']][0]
        Text['显示名称'].draw_text('')
        return Actor[kw].mouseLeave(**kwargs)


def LeftButtonPressRelease(**kwargs):
    if kwargs['picked'] in Actor.values():
        kw = [kw for kw in Actor if Actor[kw] is kwargs['picked']][0]

        if kwargs['double']:
            text = '{}-{}-{}-{}-{}-{}-{}-{}'.format(Data['长度'], Data['宽度'], Data['高度中'],
                                                    Data['高度尾'], Data['高度头'], Data['高度前'], Data['高度后'], kw)
            vmi.app.clipboard().setText(text)
            vmi.ccSaveFile_STEP(Data[kw])


def NoButtonWheel(**kwargs):
    if kwargs['picked'] in Actor.values():
        kw = [kw for kw in Actor if Actor[kw] is kwargs['picked']][0]
        Actor[kw].setOpacity(1 if Actor[kw].opacity() < 1 else 0.1)


def Input(args):
    args = dict(zip(InputKeys, args))

    for kw in args:
        try:
            args[kw] = float(args[kw])
        except:
            pass

    Data.update(args)


def Rebuild():
    W.updateInTime()
    time = QTime()
    time.start()

    global Data
    text = '输入'
    for k in InputKeys:
        text += ' ' + str(Data[k])
    print(text)

    print('{:.2f}'.format(time.elapsed() / 1e3), '开始')

    Data['左'] = np.array([1, 0, 0], np.float)
    Data['后'] = np.array([0, 1, 0], np.float)
    Data['上'] = np.array([0, 0, 1], np.float)
    Data['原点'] = np.array([0, 0, 0], np.float)

    Data['长向'] = (-1 if Data['后路侧'] > 0 else 1) * Data['左']
    Data['入旋向'] = np.cross(Data['长向'], Data['后'])

    Data['长向'] = vmi.CS4x4().rotate(-Data['横偏角'], Data['入旋向']).mvt(Data['长向'])
    Data['宽向'] = np.cross(Data['入旋向'], Data['长向'])
    Data['高向'] = Data['上']

    lh, wh, th = 0.5 * Data['长度'], 0.5 * Data['宽度'], Data['壁厚']

    # 长宽高坐标系
    cs = vmi.CS4x4(origin=Data['原点'],
                   axis0=Data['长向'],
                   axis1=Data['宽向'],
                   axis2=Data['高向'])

    # 长宽坐标
    length_width = np.array([
        [[-(lh - wh), -wh], [-lh, -0.5 * wh], [-(lh - 1.5 * wh), wh]],
        [[0, -wh], [0, 0], [0, wh]],
        [[lh - 1.5 * wh, -wh], [lh, 0.5 * wh], [lh - wh, wh]]], np.float)

    # 高坐标
    height = np.array([
        [0, Data['高度尾'], 0],
        [Data['高度前'], Data['高度中'], Data['高度后']],
        [0, Data['高度头'], 0]], np.float)

    height[0][0] = 0.5 * (height[0][1] + height[1][0])
    height[0][2] = 0.5 * (height[0][1] + height[1][2])
    height[2][0] = 0.5 * (height[2][1] + height[1][0])
    height[2][2] = 0.5 * (height[2][1] + height[1][2])

    points = {}

    # 外侧底面点
    Data['点外下'] = np.zeros((3, 3, 3), np.float)
    for j in range(3):
        for i in range(3):
            Data['点外下'][j][i] = cs.mpt([*length_width[j][i], 0])

    # 外侧顶面点
    Data['点外上'] = np.zeros((3, 3, 3), np.float)
    for j in range(3):
        for i in range(3):
            Data['点外上'][j][i] = cs.mpt([0, 0, height[j][i]], origin=Data['点外下'][j][i])

    # 内侧底面点
    Data['点内下'] = np.zeros((3, 3, 3), np.float)
    for j in range(3):
        for i in range(3):
            Data['点内下'][j][i] = cs.mpt([0, 0, th], origin=Data['点外下'][j][i])

    # 内侧顶面点
    Data['点内上'] = np.zeros((3, 3, 3), np.float)
    for j in range(3):
        for i in range(3):
            Data['点内上'][j][i] = cs.mpt([0, 0, -th], origin=Data['点外上'][j][i])

    # 内侧点向内缩进
    for inner in [Data['点内下'], Data['点内上']]:
        inner[0][0] = cs.mpt([0, th, 0], origin=inner[0][0])
        inner[0][1] = cs.mpt([th, 0, 0], origin=inner[0][1])
        inner[0][2] = cs.mpt([0, -th, 0], origin=inner[0][2])
        inner[1][0] = cs.mpt([0, th, 0], origin=inner[1][0])
        inner[1][2] = cs.mpt([0, -th, 0], origin=inner[1][2])
        inner[2][0] = cs.mpt([0, th, 0], origin=inner[2][0])
        inner[2][1] = cs.mpt([-th, 0, 0], origin=inner[2][1])
        inner[2][2] = cs.mpt([0, -th, 0], origin=inner[2][2])

    # 显示点
    pts = []
    for pt in [Data['点外下'], Data['点外上'], Data['点内下'], Data['点内上']]:
        for j in range(3):
            for i in range(3):
                pts.append(vmi.ccVertex(pt[j][i]))
    Data['点'] = vmi.ccBoolean_Union(pts)

    # 切向量
    tg_xy = np.array(
        [[[-2, 0], [0, 1], [2, 1]],
         [[0, 1], [0, 1], [0, 1]],
         [[2, 1], [0, 1], [-2, 0]]])

    tg = np.zeros((3, 3, 3), np.float)
    for j in range(3):
        for i in range(3):
            tg[j][i] = cs.mvt([*tg_xy[j][i], 0])

    # 切向量约束状态
    tgflag = [[True, True, True], [False, True, False], [True, True, True]]

    for l in ['外', '内']:
        for k in ['下', '上']:
            Data['宽边' + l + k] = [0] * 3

            for j in range(3):
                if l == '外':
                    curve = vmi.ccBSpline(Data['点' + l + k][j], tg[j], tgflag[j])
                else:
                    curve = vmi.ccSegment(Data['点' + l + k][j][0], Data['点' + l + k][j][2])

                Data['宽边' + l + k][j] = vmi.ccEdge(curve)

                kw = '宽边' + l + ['尾', '中', '头'][j] + k
                Data[kw] = Data['宽边' + l + k][j]

        for k in ['下', '上']:
            Data['端面' + l + k] = vmi.ccLoft(Data['宽边' + l + k])

            endpts = [vmi.ccEdge_FirstLastPoint(Data['宽边' + l + k][i]) for i in [0, 2]]
            for edge in vmi.ccExplore_Recursive(Data['端面' + l + k], vmi.TopAbs_EDGE):
                pts = vmi.ccEdge_FirstLastPoint(edge)
                if pts[0] in [endpts[0][0], endpts[1][0]] and pts[1] in [endpts[0][0], endpts[1][0]]:
                    Data['长边' + l + k + '前'] = edge
                elif pts[0] in [endpts[0][1], endpts[1][1]] and pts[1] in [endpts[0][1], endpts[1][1]]:
                    Data['长边' + l + k + '后'] = edge

        Data['端面' + l + '右'] = vmi.ccLoft([Data['宽边' + l + '下'][0], Data['宽边' + l + '上'][0]], ruled=True)
        Data['端面' + l + '左'] = vmi.ccLoft([Data['宽边' + l + '下'][2], Data['宽边' + l + '上'][2]], ruled=True)

        for j in ['前', '后']:
            Data['端面' + l + j] = vmi.ccLoft([Data['长边' + l + '下' + j], Data['长边' + l + '上' + j]], ruled=True)

        Data['实体' + l] = vmi.ccSolid(vmi.ccSew([Data['端面' + l + i] for i in ['下', '上', '右', '左', '前', '后']]))

    solid = vmi.ccBoolean_Difference([Data['实体外'], Data['实体内']])

    print('{:.2f}'.format(time.elapsed() / 1e3), '创建外壳')

    pts = [[cs.mpt([0, 0, 0], origin=Data['点内下'][0][0]),
            cs.mpt([0, 0, 0], origin=Data['点内上'][0][0]),
            cs.mpt([0.25 * wh - 0.5 * th, 0, 0], origin=Data['点内下'][1][0]),
            cs.mpt([0.25 * wh - 0.5 * th, 0, 0], origin=Data['点内上'][1][0]),
            cs.mpt([0.25 * wh + 0.5 * th, 0, 0], origin=Data['点内下'][1][0]),
            cs.mpt([0.25 * wh + 0.5 * th, 0, 0], origin=Data['点内上'][1][0]),
            cs.mpt([0, 0, 0], origin=Data['点内下'][2][0]),
            cs.mpt([0, 0, 0], origin=Data['点内上'][2][0])],
           [cs.mpt([0, 0, 0], origin=Data['点内下'][0][2]),
            cs.mpt([0, 0, 0], origin=Data['点内上'][0][2]),
            cs.mpt([-0.25 * wh - 0.5 * th, 0, 0], origin=Data['点内下'][1][2]),
            cs.mpt([-0.25 * wh - 0.5 * th, 0, 0], origin=Data['点内上'][1][2]),
            cs.mpt([-0.25 * wh + 0.5 * th, 0, 0], origin=Data['点内下'][1][2]),
            cs.mpt([-0.25 * wh + 0.5 * th, 0, 0], origin=Data['点内上'][1][2]),
            cs.mpt([0, 0, 0], origin=Data['点内下'][2][2]),
            cs.mpt([0, 0, 0], origin=Data['点内上'][2][2])]]

    cut = [[vmi.ccWire(vmi.ccSegments([pts[0][0], pts[0][1], pts[0][3], pts[0][2]], True)),
            vmi.ccWire(vmi.ccSegments([pts[0][7], pts[0][6], pts[0][4], pts[0][5]], True))],
           [vmi.ccWire(vmi.ccSegments([pts[1][0], pts[1][1], pts[1][3], pts[1][2]], True)),
            vmi.ccWire(vmi.ccSegments([pts[1][7], pts[1][6], pts[1][4], pts[1][5]], True))]]
    cut = [[vmi.ccPrism(vmi.ccFace(c), [-i for i in Data['宽向']], th) for c in cut[0]],
           [vmi.ccPrism(vmi.ccFace(c), Data['宽向'], th) for c in cut[1]]]

    solid = vmi.ccBoolean_Difference([solid, *cut[0], *cut[1]])

    porous = vmi.ccBoolean_Union(cut[0] + cut[1])

    print('{:.2f}'.format(time.elapsed() / 1e3), '创建前后窗')

    pts = [cs.mpt([0, 0, -th], Data['点内下'][0][0]),
           cs.mpt([0, 0, -th], Data['点内下'][0][2]),
           cs.mpt([0.25 * wh - 0.5 * th, 0, -th], Data['点内下'][1][0]),
           cs.mpt([-0.25 * wh - 0.5 * th, 0, -th], Data['点内下'][1][2]),
           cs.mpt([0.25 * wh + 0.5 * th, 0, -th], Data['点内下'][1][0]),
           cs.mpt([-0.25 * wh + 0.5 * th, 0, -th], Data['点内下'][1][2]),
           cs.mpt([0, 0, -th], Data['点内下'][2][0]),
           cs.mpt([0, 0, -th], Data['点内下'][2][2])]

    cut = [vmi.ccSegments([pts[0], pts[1], pts[3], pts[2]], True),
           vmi.ccSegments([pts[4], pts[5], pts[7], pts[6]], True)]
    cut = [vmi.ccWire([vmi.ccEdge(c) for c in cut[0]]),
           vmi.ccWire([vmi.ccEdge(c) for c in cut[1]])]
    cut = [vmi.ccPrism(vmi.ccFace(c), Data['高向'], 2 * Data['高度中']) for c in cut]

    if not Data['植骨窗'] > 0:
        porous = vmi.ccBoolean_Union([vmi.ccBoolean_Intersection([solid, *cut]), porous])
        print('{:.2f}'.format(time.elapsed() / 1e3), '创建植骨窗')

    solid = vmi.ccBoolean_Difference([solid, *cut])

    print('{:.2f}'.format(time.elapsed() / 1e3), '创建上下窗')

    if Data['植骨孔'] > 0:
        hmin = min(height[2][0] - th, height[2][1], height[2][2] - th)
        r = min(0.5 * (hmin - th), wh - th) - 0.5
        Data['植骨孔直径'] = 2 * r
        cut = cs.mpt([0, 0, (hmin + th) / 2], Data['点外下'][1][1])
        cut = vmi.ccCylinder(r, Data['长度'], cut, Data['长向'], Data['高向'])
        solid = vmi.ccBoolean_Difference([solid, cut])

        print('{:.2f}'.format(time.elapsed() / 1e3), '创建植骨孔')

    l, w, h = wh - th, 0.5 * Data['夹持宽度'], 0.5 * Data['夹持高度']
    pts = [cs.mpt([-lh, -w - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-lh, -(w + wh) - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-(lh - l), -w - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-(lh - l), -(w + wh) - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-lh, w - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-lh, w + wh - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-(lh - l), w - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点']),
           cs.mpt([-(lh - l), w + wh - 0.25 * wh, 0.5 * Data['高度尾'] - h], Data['原点'])]

    box = [vmi.ccWire(vmi.ccSegments([pts[0], pts[1], pts[3], pts[2]], True)),
           vmi.ccWire(vmi.ccSegments([pts[4], pts[5], pts[7], pts[6]], True))]
    box = [vmi.ccPrism(vmi.ccFace(b), Data['高向'], 2 * h) for b in box]
    box = vmi.ccBoolean_Union([box[0], box[1]])
    solid = vmi.ccBoolean_Difference([solid, box])

    print('{:.2f}'.format(time.elapsed() / 1e3), '创建夹持槽')

    n = int(vmi.ccLength(Data['宽边外下'][0]) / Data['敲击带半径'])
    pts = [np.array(vmi.ccUniformPts_Edge(Data['宽边外下'][0], n + 1)),
           np.array(vmi.ccUniformPts_Edge(Data['宽边外上'][0], n + 1))]
    for i in range(2, n + 1 - 2, 3):
        c = [cs.mpt([0, 0, 0.5 * Data['高度尾']], pts[0][i - 2]),
             cs.mpt([0, 0, 0.5 * Data['高度尾']], pts[0][i]),
             cs.mpt([0, 0, 0.5 * Data['高度尾']], pts[0][i + 2])]

        cylinders = []
        if not vmi.ccInside_Pt(box, c[0]) and not vmi.ccInside_Pt(box, c[1]) and not vmi.ccInside_Pt(box, c[2]):
            h = np.linalg.norm(pts[0][i] - pts[1][i]) - 2 * th
            o = cs.mpt([0, 0, th], pts[0][i])
            cylinders.append(vmi.ccCylinder(Data['敲击带半径'], h, o, Data['高向'], Data['长向']))

        solid = vmi.ccBoolean_Union([solid, *cylinders])

    print('{:.2f}'.format(time.elapsed() / 1e3), '创建敲击带')

    Data['实体'], Data['多孔'] = solid, porous
    Actor['实体'].setData(solid)
    Actor['多孔'].setData(porous)

    print('{:.2f}'.format(time.elapsed() / 1e3), '离散化')


InputKeys = ['后路侧', '横偏角',
             '长度', '宽度',
             '高度中', '高度尾', '高度头', '高度前', '高度后',
             '壁厚',
             '植骨窗', '植骨孔',
             '夹持宽度', '夹持高度',
             '敲击带半径']

Data = {'后路侧': 1, '横偏角': 15,
        '长度': 32, '宽度': 12,
        '高度中': 14, '高度尾': 12, '高度头': 12, '高度前': 13, '高度后': 13,
        '壁厚': 1.5,
        '植骨窗': 0, '植骨孔': 1, '植骨孔直径': 0.5,
        '夹持宽度': 3, '夹持高度': 3,
        '敲击带半径': 0.3}

Input(sys.argv[1:])

Actor, Text = {}, {}


def info():
    text = '输入 (mm)' + '\n\n'
    text += '后路侧: ' + ('左' if Data['后路侧'] > 0 else '右') + '\n'
    text += '横偏角: ' + str(Data['横偏角']) + '°' + '\n'
    text += '长度: ' + str(Data['长度']) + '\n'
    text += '宽度: ' + str(Data['宽度']) + '\n'
    text += '高度: ' + '中 ' + str(Data['高度中']) + \
            ' 尾 ' + str(Data['高度尾']) + ' 头 ' + str(Data['高度头']) + \
            ' 前 ' + str(Data['高度前']) + ' 后 ' + str(Data['高度后']) + '\n'
    text += '壁厚: ' + str(Data['壁厚']) + '\n'
    text += '植骨窗: ' + ('启用' if Data['植骨窗'] > 0 else '禁用') + '\n'
    text += '植骨孔: ' + ('启用' if Data['植骨孔'] > 0 else '禁用') + ' 直径: {:.2f}'.format(Data['植骨孔直径']) + '\n'
    text += '夹持槽: ' + '宽度 ' + str(Data['夹持宽度']) + ' 高度 ' + str(Data['夹持高度']) + '\n'
    text += '敲击带: ' + '半径 ' + str(Data['敲击带半径']) + '\n'

    print(text)


Actor['点'] = vmi.PolyActor(W, name='点', color=[0.2, 0.6, 1], point_size=8, pickable=True)
Actor['点'].mouse['NoButton']['Enter'] = NoButtonEnter
Actor['点'].mouse['NoButton']['Leave'] = NoButtonLeave

for l in ['外', '内']:
    for k in ['下', '上']:
        for j in ['尾', '中', '头']:
            kw = '宽边' + l + j + k
            Actor[kw] = vmi.PolyActor(W, name=kw, color=[1, 0.2, 0.2], line_width=4, pickable=True)
            Actor[kw].mouse['NoButton']['Enter'] = NoButtonEnter
            Actor[kw].mouse['NoButton']['Leave'] = NoButtonLeave

    for k in ['下', '上']:
        for kw in ['长边' + l + k + kk for kk in ['前', '后']]:
            Actor[kw] = vmi.PolyActor(W, name=kw, color=[1, 0.2, 0.2], line_width=4, pickable=True)
            Actor[kw].mouse['NoButton']['Enter'] = NoButtonEnter
            Actor[kw].mouse['NoButton']['Leave'] = NoButtonLeave

kw = '实体'
Actor[kw] = vmi.PolyActor(W, name=kw, color=[0.9, 0.9, 0.9], pickable=True)
Actor[kw].mouse['NoButton']['Enter'] = NoButtonEnter
Actor[kw].mouse['NoButton']['Leave'] = NoButtonLeave
Actor[kw].mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
Actor[kw].mouse['NoButton']['Wheel'] = NoButtonWheel

kw = '多孔'
Actor[kw] = vmi.PolyActor(W, name=kw, color=[0.9, 0.5, 0.5], pickable=True)
Actor[kw].mouse['NoButton']['Enter'] = NoButtonEnter
Actor[kw].mouse['NoButton']['Leave'] = NoButtonLeave
Actor[kw].mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
Actor[kw].mouse['NoButton']['Wheel'] = NoButtonWheel

Text['显示名称'] = vmi.TextBox(W, text_align=Qt.AlignCenter, size=[0.1, 0.05], pos=[0, 1], anchor=[0, 1])

T = Thread(target=Rebuild)
T.start()
vmi.appwait(T)

info()

W.setCamera_Coronal()
W.setCamera_Reverse()
W.setCamera_FitAll()
vmi.appexec(W)



import pathlib
import tempfile

import vmi
import nibabel as nib
import numpy as np
import pydicom
import vtk
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from numba import jit
import pandas as pd
import os
import pymongo
import gridfs
import pickle
import hashlib
import threading

from typing import List, Optional
from typing import Dict, Any

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from PySide2.QtWidgets import QInputDialog

dic_list = ['髂翼最高点',
            '髂前上棘',
            '髂前下棘',
            '耻骨结节',
            '坐骨结节',
            '髂后上棘',
            '坐骨棘',
            '泪滴']

color_map = {
    '髂翼最高点': [128, 0, 0],
    '髂前上棘': [128, 128, 0],
    '髂前下棘': [128, 0, 128],
    '耻骨结节': [0, 255, 128],
    '坐骨结节': [0, 128, 128],
    '髂后上棘': [0, 0, 255],
    '坐骨棘': [255, 128, 255],
    '泪滴': [0, 255, 255],
}

# mark_points = {
#     '髂前上棘': [],
#     '耻骨结节': [],
#     '髂翼最高点': [],
#     '坐骨结节': [],
#     '髂后上棘': [],
#     '泪滴': []}

overlay_chart = ['正视', '侧视', '俯视']

SizePolicyMinimum = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)


def upload(image: vtk.vtkImageData, kw_value: Dict[str, Any]):
    if flag == 1:
        data_db: pymongo.collection.Collection = client.THA_DB.prePelvis
        file_db = gridfs.GridFS(client.THA_DB, collection='prePelvis')

    elif flag == 2:
        data_db: pymongo.collection.Collection = client.THA_DB.postPelvis
        file_db = gridfs.GridFS(client.THA_DB, collection='postPelvis')

    # 删除已有数据
    uid = {'SeriesInstanceUID': kw_value['SeriesInstanceUID']}

    for one in file_db.find(uid):
        file_db.delete(one._id)

    for one in data_db.find(uid):
        data_db.delete_one(one)

    # 图像序列化为文件，计算特征码，上传文件项
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.nii'
        vmi.imSaveFile_NIFTI(image, str(p))
        image_bytes = p.read_bytes()
        file_id = file_db.put(image_bytes,
                              fileType='.nii',
                              scalarType=image.GetScalarTypeAsString(),
                              SeriesInstanceUID=kw_value['SeriesInstanceUID'])

    # 检查上传结果是否与本地一致
    if file_db.find_one({'_id': file_id}).md5 != hashlib.md5(image_bytes).hexdigest():
        file_db.delete(file_id)
        return print('远端MD5校验失败 {}'.format(repr(kw_value)))

    # 上传数据项
    data_db.insert_one(kw_value)


def input_dicom():
    global series_list, dcmdir
    dcmdir = vmi.askDirectory('DICOM')

    if dcmdir is not None:
        series_list = vmi.sortSeries(dcmdir)
        input_series()


def input_series():
    global series_dict, series, voi_image
    series = vmi.askSeries(series_list)
    if series is not None:
        series_dict = series.toKeywordValue()
        voi_image = series.read()
        voi_image.SetOrigin([0, 0, 0])
        input_image(voi_image)


def input_image(image: vtk.vtkImageData):
    global center_pts, mark_hash, front_img, side_img, top_img

    voi_volume.setData(image)
    voi_volume.setOpacity_Threshold(threshold)
    voi_view.setCamera_Coronal()
    voi_view.setCamera_FitAll()

    for j in center_props:
        j.setData(vmi.pdSphere(0.01, [0, 0, 0]))

    for j in image_props:
        j.setData(vmi.pdSphere(0.01, [0, 0, 0]))

    mark_hash = []
    center_pts = []
    front_img, side_img, top_img = None, None, None
    front_img = show_Front_view(voi_image)

    image_data.setData(front_img)
    image_data.setColorWindow_Auto()
    image_view.setCamera_FitAll()

    voi_view.updateAtOnce()
    vmi.appupdate()


def bind_props(axis):
    global image_props
    x_offset = vmi.imExtent(voi_image)[1] - vmi.imExtent(voi_image)[0] + 1
    y_offset = vmi.imExtent(voi_image)[3] - vmi.imExtent(voi_image)[2] + 1
    z_offset = vmi.imExtent(voi_image)[5] - vmi.imExtent(voi_image)[4] + 1
    x_spacing = vmi.imSpacing(voi_image)[0]
    y_spacing = vmi.imSpacing(voi_image)[1]
    z_spacing = vmi.imSpacing(voi_image)[2]

    image_front_cs = vmi.CS4x4(axis0=[1, 0, 0],
                               axis1=[0, 0, -1],
                               axis2=[0, 1, 0],
                               origin=[0, 0, z_offset * z_spacing])
    image_side_cs = vmi.CS4x4(axis0=[0, -1, 0],
                              axis1=[0, 0, -1],
                              axis2=[1, 0, 0],
                              origin=[0, y_offset * y_spacing, z_offset * z_spacing])
    image_top_cs = vmi.CS4x4(axis0=[1, 0, 0],
                             axis1=[0, 1, 0],
                             axis2=[0, 0, 1],
                             origin=[0, 0, 0])

    ith = -1

    for i in range(len(center_pts)):
        if i % 2 == 0:
            ith += 1
            if ith >= 5:
                ith = 5
        rbg = np.array(color_map[dic_list[ith]]) / 255
        image_props[i].setColor(rbg)
        if axis == 1:
            image_props[i].setData(vmi.pdSphere(radius, [image_front_cs.inv().mpt(center_pts[i])[0],
                                                         image_front_cs.inv().mpt(center_pts[i])[1], 0]))
        elif axis == 2:
            side_pt = [image_side_cs.inv().mpt(center_pts[i])[0], image_side_cs.inv().mpt(center_pts[i])[1], 0]
            image_props[i].setData(vmi.pdSphere(radius, side_pt))
        elif axis == 0:
            image_props[i].setData(vmi.pdSphere(radius, [image_top_cs.inv().mpt(center_pts[i])[0],
                                                         image_top_cs.inv().mpt(center_pts[i])[1], 0]))
        image_props[i].alwaysOnTop()


def _dialog_ok_cancel(widgets, title, ok_cancel=True) -> QDialog:
    dialog = QDialog()
    dialog.setSizePolicy(SizePolicyMinimum)
    dialog.setWindowTitle(title)
    dialog.setLayout(QVBoxLayout())
    dialog.setMinimumWidth(800)

    for w in widgets:
        w.setParent(dialog)
        dialog.layout().addWidget(w)

    if ok_cancel:
        button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
        button.setSizePolicy(SizePolicyMinimum)
        button.accepted.connect(dialog.accept)
        button.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button)
    return dialog


# 访问数据列表
def load_by_list():
    global case_uid, case_info, case_image_data, voi_image, mark_hash, center_props, center_pts, front_img, side_img, top_img

    data_db: pymongo.collection.Collection = client.testDB.spinePelvis
    file_db = gridfs.GridFS(client.testDB, collection='spinePelvis')
    markData_db: pymongo.collection.Collection = client.testDB.spinePelvisMF

    tree = QTreeWidget()
    tree.setColumnCount(5)
    tree.setHeaderLabels(['姓名', '年龄', '模态', '层数', '序列'])

    for x in data_db.find({"SeriesInstanceUID": {'$exists': True}}).sort(
            "PatientName", pymongo.ASCENDING):
        if not markData_db.find_one({'SeriesInstanceUID': x["SeriesInstanceUID"]}):
            item = QTreeWidgetItem()
            item.setText(0, str(x["PatientName"]))
            item.setText(1, str(x["PatientAge"]))
            item.setText(2, str(x["Modality"]))
            item.setText(3, str(x["Slices"]))
            item.setText(4, str(x["SeriesInstanceUID"]))
            tree.addTopLevelItem(item)

    d = _dialog_ok_cancel([tree], "list")
    if d.exec_() == QDialog.Accepted:
        case_uid = tree.currentItem().text(4)
        search_dict = {'SeriesInstanceUID': case_uid}

        if not file_db.exists(search_dict):
            return 'case_uid not found'
        else:
            case_info = data_db.find_one({'SeriesInstanceUID': case_uid})
            case_image_data = file_db.find_one(search_dict)

            # 清空所有点
            for i in center_props:
                i.setData(vmi.pdSphere(0.01, [0, 0, 0]))

            mark_hash = []
            center_pts = []
            front_img, side_img, top_img = None, None, None

            for j in image_props:
                j.setData(vmi.pdSphere(0.01, [0, 0, 0]))
            # 反序列化并将数据传至voi_volum
            voi_image = pickle.loads(case_image_data.read())
            voi_image.SetOrigin([0, 0, 0])

            front_img = show_Front_view(voi_image)

            voi_volume.setData(voi_image)
            voi_volume.setOpacity_Threshold(threshold)
            voi_view.setCamera_FitAll()
            voi_view.setCamera_Coronal()

            image_data.setData(front_img)
            image_data.setColorWindow_Auto()
            image_view.setCamera_FitAll()


# 上传文件
def output_upload():
    global case_uid
    if flag == 0:
        markData_db: pymongo.collection.Collection = client.testDB.spinePelvisMF
        markFile_db = gridfs.GridFS(client.testDB, collection='spinePelvisMF')

    elif flag == 1:
        markData_db: pymongo.collection.Collection = client.THA_DB.prePelvisMF
        markFile_db = gridfs.GridFS(client.THA_DB, collection='prePelvisMF')

        if series:
            tags = series.toKeywordValue()
            case_uid = tags['SeriesInstanceUID']
            upload(voi_image, tags)

    elif flag == 2:
        markData_db: pymongo.collection.Collection = client.THA_DB.postPelvisMF
        markFile_db = gridfs.GridFS(client.THA_DB, collection='postPelvisMF')

        if series:
            tags = series.toKeywordValue()
            case_uid = tags['SeriesInstanceUID']
            upload(voi_image, tags)

    if center_pts:
        mark_dict = {
            '髂翼最高点': [center_pts[0], center_pts[1]],
            '髂前上棘': [center_pts[2], center_pts[3]],
            '髂前下棘': [center_pts[4], center_pts[5]],
            '耻骨结节': [center_pts[6], center_pts[7]],
            '坐骨结节': [center_pts[8], center_pts[9]],
            '髂后上棘': [center_pts[10], center_pts[11]],
            '坐骨棘': [center_pts[12], center_pts[13]],
            '泪滴': [center_pts[14], center_pts[15]],
        }

        mark_bytes = pickle.dumps(mark_dict)
        mark_hash.append(hashlib.md5(mark_bytes).hexdigest())

        # 上传文件，避免重复上传完全一样的文件
        if not markFile_db.exists({'markData': mark_hash}):
            markFile_db.put(mark_bytes, fileType='.json')

            if not markData_db.find_one({'SeriesInstanceUID': case_uid}):
                markData_db.insert_one({'SeriesInstanceUID': case_uid})

            # 替换旧数据,增加marked数据的md5
            unmarked_data_dict = markData_db.find_one({'SeriesInstanceUID': case_uid})
            marked_data_dict = {**unmarked_data_dict, 'markData': mark_hash}
            markData_db.find_one_and_replace(unmarked_data_dict, marked_data_dict)

            data_id = mark_hash[-1]

            # 记录上传的id
            if vmi.askYesNo('上传成功 {} 复制到剪贴板？'.format(str(data_id))):
                vmi.app.clipboard().setText(str(data_id))


def _Overlay(image_data, axis):
    np_data = vmi.imArray_VTK(image_data)
    np_data[np_data < 100] = 0
    if axis == 0:
        overlay_data = np_data.sum(axis=0)
        return vmi.imVTK_Array(overlay_data, origin=[0, 0, 0],
                               spacing=[vmi.imSpacing(image_data)[1], vmi.imSpacing(image_data)[0],
                                        vmi.imSpacing(image_data)[2]])
    elif axis == 1:
        overlay_data = np.flip(np.flip(np.sum(np_data, axis)), axis=1) / 100
        return vmi.imVTK_Array(overlay_data, origin=[0, 0, 0],
                               spacing=[vmi.imSpacing(image_data)[0], vmi.imSpacing(image_data)[2],
                                        vmi.imSpacing(image_data)[1]])
    elif axis == 2:
        overlay_data = np.fliplr(np.flip(np.flip(np.sum(np_data, axis)), axis=1) / 100)
    return vmi.imVTK_Array(overlay_data, origin=[0, 0, 0],
                           spacing=[vmi.imSpacing(image_data)[0], vmi.imSpacing(image_data)[2],
                                    vmi.imSpacing(image_data)[1]])


def show_Front_view(image_data, axis=1):
    fount_img = _Overlay(image_data, axis)
    return fount_img


def show_Side_view(image_data, axis=2):
    side_img = _Overlay(image_data, axis)
    return side_img


def show_Top_view(image_data, axis=0):
    top_img = _Overlay(image_data, axis)
    return top_img


# 查看已上传数据
def check_upload():
    global case_uid, case_info, case_image_data, voi_image, mark_hash, center_pts, front_img, side_img, top_img

    if flag == 0:
        data_db: pymongo.collection.Collection = client.testDB.spinePelvis
        file_db = gridfs.GridFS(client.testDB, collection='spinePelvis')
        markData_db: pymongo.collection.Collection = client.testDB.spinePelvisMF
        markFile_db = gridfs.GridFS(client.testDB, collection='spinePelvisMF')

    elif flag == 1:
        data_db: pymongo.collection.Collection = client.THA_DB.prePelvis
        file_db = gridfs.GridFS(client.THA_DB, collection='prePelvis')
        markData_db: pymongo.collection.Collection = client.THA_DB.prePelvisMF
        markFile_db = gridfs.GridFS(client.THA_DB, collection='prePelvisMF')

    elif flag == 2:
        data_db: pymongo.collection.Collection = client.THA_DB.postPelvis
        file_db = gridfs.GridFS(client.THA_DB, collection='postPelvis')
        markData_db: pymongo.collection.Collection = client.THA_DB.postPelvisMF
        markFile_db = gridfs.GridFS(client.THA_DB, collection='postPelvisMF')

    tree = QTreeWidget()
    tree.setColumnCount(5)
    tree.setHeaderLabels(['姓名', '年龄', '模态', '层数', '序列'])

    for x in markData_db.find({"SeriesInstanceUID": {'$exists': True}}):
        if data_db.find_one({"SeriesInstanceUID": x["SeriesInstanceUID"]}):
            k = data_db.find_one({"SeriesInstanceUID": x["SeriesInstanceUID"]})

            item = QTreeWidgetItem()
            item.setText(0, str(k["PatientName"]))
            item.setText(1, str(k["PatientAge"]))
            item.setText(2, str(k["Modality"]))
            item.setText(3, str(k["slices"]))
            item.setText(4, str(k["SeriesInstanceUID"]))
            tree.addTopLevelItem(item)

    d = _dialog_ok_cancel([tree], "list")
    if d.exec_() == QDialog.Accepted:
        case_uid = tree.currentItem().text(4)
        search_dict = {'SeriesInstanceUID': case_uid}
        if not file_db.exists(search_dict):
            return 'case_uid not found'
        else:
            case_info = data_db.find_one({'SeriesInstanceUID': case_uid})
            case_image_data = file_db.find_one(search_dict)

            # 反序列化并将数据传至voi_volum
            voi_image = pickle.loads(case_image_data.read())
            voi_image.SetOrigin([0, 0, 0])
            voi_volume.setData(voi_image)
            voi_volume.setOpacity_Threshold(threshold)
            voi_view.setCamera_FitAll()
            voi_view.setCamera_Coronal()

            mark_hash = markData_db.find_one({"SeriesInstanceUID": case_uid})["markData"]
            pts_info = markFile_db.find_one({"md5": mark_hash[-1]})
            pts_dict = pickle.loads(pts_info.read())

            center_pts = []
            front_img, side_img, top_img = None, None, None
            front_img = show_Front_view(voi_image)
            image_data.setData(front_img)
            image_data.setColorWindow_Auto()
            image_view.setCamera_FitAll()
            for j in image_props:
                j.setData(vmi.pdSphere(0.01, [0, 0, 0]))
            for name in pts_dict:
                center_pts.append(pts_dict[name][0])
                center_pts.append(pts_dict[name][1])

            render_all_pts()
            bind_props(axis=1)


# 添加并更新所有标记点
def render_all_pts():
    ith = -1
    global center_props, center_pts
    for i in range(len(center_pts)):
        if i % 2 == 0:
            ith += 1
            rbg = np.array(color_map[dic_list[ith]]) / 255
        center_props[i].setColor(rbg)
        center_props[i].setData(vmi.pdSphere(radius, center_pts[i]))
    voi_view.updateAtOnce()


def LeftButtonPressRelease(**kwargs):
    global center_pts, image_display_method
    if kwargs['picked'] in [voi_view]:
        if kwargs['double'] is True:
            if len(center_pts) < 16:
                center_pts.append(voi_view.pickPt_Cell())
                render_all_pts()
                bind_props(axis=image_display_method)
            else:
                pass
    else:
        image_view.pickPt_FocalPlane()


def LeftButtonPress(**kwargs):
    global visable, image_data, front_img, side_img, top_img, image_view, image_display_method, center_pts, flag
    if kwargs['picked'] is features_box_list[0]:
        output_upload()

    elif kwargs['picked'] is features_box_list[1]:
        load_by_list()

    elif kwargs['picked'] is features_box_list[3]:
        # 显示/隐藏体数据
        visable = not (visable)
        voi_volume.setVisible(visable)
        if visable:
            features_box_list[3].draw_text(text='显示体数据')
        else:
            features_box_list[3].draw_text(text='隐藏体数据')

    elif kwargs['picked'] is overlay_chart_text_box_list[0]:
        if front_img is None:
            front_img = show_Front_view(voi_image)
        image_data.setData(front_img)
        bind_props(axis=1)
        image_data.setColorWindow_Auto()
        image_view.setCamera_FitAll()
        image_display_method = 1

    elif kwargs['picked'] is overlay_chart_text_box_list[1]:
        if side_img is None:
            side_img = show_Side_view(voi_image)

        image_data.setData(side_img)
        bind_props(axis=2)
        image_view.setCamera_FitAll()
        image_data.setColorWindow_Auto()
        image_display_method = 2

    elif kwargs['picked'] is overlay_chart_text_box_list[2]:
        if top_img is None:
            top_img = show_Top_view(voi_image)
        image_data.setData(top_img)
        bind_props(axis=0)
        image_view.setCamera_FitAll()
        image_data.setColorWindow_Auto()
        image_display_method = 0

    elif kwargs['picked'] is features_box_list[4]:
        check_upload()

    elif kwargs['picked'] is features_box_list[5]:
        flag = 1
        d = vmi.askYesNo('是否确认上传术前数据？')
        if d:
            output_upload()

    elif kwargs['picked'] is features_box_list[6]:
        flag = 2
        d = vmi.askYesNo('是否确认上传术后数据？')
        if d:
            output_upload()

    elif kwargs['picked'] is features_box_list[7]:
        flag = 1
        check_upload()

    elif kwargs['picked'] is features_box_list[8]:
        flag = 2
        check_upload()


def LeftButtonPressMove(**kwargs):
    if kwargs['picked'] in center_props:
        i = center_props.index(kwargs['picked'])
        #     center_pts[i] += voi_view.pickVt_FocalPlane()
        #     center_props[i].setData(vmi.pdSphere(radius, center_pts[i]))
        # else:
        #     voi_view.mouseRotateFocal(**kwargs)

        center_pts[i] += voi_view.pickVt_FocalPlane()

        center_props[i].setData(vmi.pdSphere(radius, center_pts[i]))
        bind_props(axis=image_display_method)

    else:
        voi_view.mouseRotateFocal(**kwargs)


def NoButtonWheel(**kwargs):
    if kwargs['picked'] is features_box_list[2]:
        global threshold
        threshold = min(max(threshold + 10 * kwargs['delta'], -1000), 3000)
        features_box_list[2].draw_text('阈值：{} HU'.format(threshold))
        voi_volume.setOpacityScalar({threshold - 1: 0, threshold: 1})


class Main(QScrollArea):
    def __init__(self):
        QScrollArea.__init__(self)
        brush = QBrush(QColor(255, 255, 255, 255))
        palette = QPalette()
        palette.setBrush(QPalette.Active, QPalette.Base, brush)
        palette.setBrush(QPalette.Active, QPalette.Window, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Base, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Window, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Base, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Window, brush)
        self.setPalette(palette)
        self.setWidgetResizable(True)

    def closeEvent(self, ev: QCloseEvent):
        ev.accept() if vmi.askYesNo('退出程序？') else ev.ignore()


def return_globals():
    return globals()


if __name__ == '__main__':
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)

    visable = True
    threshold = 200
    image_display_method = 1
    front_img, side_img, top_img = None, None, None
    radius = 2.5
    center_props = []
    center_pts = []

    case_uid = []
    case_info = []
    case_image_data = []
    image_props = []
    mark_hash = []

    voi_image = []
    dcmdir, series, series_list = [], [], []
    flag = 0

    voi_view = vmi.View()

    for i in range(16):
        center_props.append(vmi.PolyActor(voi_view, color=[0, 0, 0], pickable=True, always_on_top=True))
    voi_volume = vmi.ImageVolume(voi_view, pickable=True)

    voi_view.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]
    voi_view.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

    textBox_list = []
    for i, key in enumerate(color_map):
        if i % 2 == 0:

            textBox_list.append(vmi.TextBox(voi_view, back_color=QColor.fromRgbF([np.array(color_map[key]) / 255][0][0],
                                                                                 [np.array(color_map[key]) / 255][0][1],
                                                                                 [np.array(color_map[key]) / 255][0][
                                                                                     2]),
                                            size=[0.02, 0.04], pos=[0, i // 2 * 0.06], anchor=[0, 0]))
            textBox_list.append(vmi.TextBox(voi_view, text=key, size=[0.1, 0.04], pos=[0.02, i // 2 * 0.06],
                                            anchor=[0, 0]))
        elif i % 2 == 1:
            textBox_list.append(vmi.TextBox(voi_view, back_color=QColor.fromRgbF([np.array(color_map[key]) / 255][0][0],
                                                                                 [np.array(color_map[key]) / 255][0][1],
                                                                                 [np.array(color_map[key]) / 255][0][
                                                                                     2]),
                                            size=[0.02, 0.04], pos=[0.12, i // 2 * 0.06], anchor=[0, 0]))
            textBox_list.append(vmi.TextBox(voi_view, text=key, size=[0.1, 0.04], pos=[0.14, i // 2 * 0.06],
                                            anchor=[0, 0]))

    features_box_list = [vmi.TextBox(voi_view, pickable=True, text='上传标记数据', size=[0.12, 0.04], pos=[0.12, 0.3],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='下载原始数据', size=[0.12, 0.04], pos=[0, 0.3],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='阈值：200 HU', size=[0.12, 0.04], pos=[0, 0.24],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='显示体数据', size=[0.12, 0.04], pos=[0.12, 0.24],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='查看已上传数据', size=[0.12, 0.04], pos=[0, 0.36],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='上传术前数据', size=[0.12, 0.04], pos=[0, 0.3],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='上传术后数据', size=[0.12, 0.04], pos=[0.12, 0.3],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='查看术前数据', size=[0.12, 0.04], pos=[0, 0.36],
                                     anchor=[0, 0]),
                         vmi.TextBox(voi_view, pickable=True, text='查看术后数据', size=[0.12, 0.04], pos=[0.12, 0.36],
                                     anchor=[0, 0])
                         ]

    features_box_list[0].setVisible(False)
    features_box_list[1].setVisible(False)
    features_box_list[4].setVisible(False)

    for box in features_box_list:
        box.mouse['LeftButton']['PressRelease'] = [LeftButtonPress]

    features_box_list[2].mouse['NoButton']['Wheel'] = [NoButtonWheel]

    image_view = vmi.View()
    image_data = vmi.ImageSlice(image_view, pickable=True)
    for i in range(16):
        image_props.append(vmi.PolyActor(image_view, color=[0, 0, 0], pickable=True, always_on_top=True))
    overlay_chart_text_box_list = []
    overlay_chart_text_box = vmi.TextBox(image_view, text='叠加图', size=[0.24, 0.04], pos=[0, 0],
                                         anchor=[0, 0])

    for i, text in enumerate(overlay_chart):
        overlay_chart_text_box_list.append(
            vmi.TextBox(image_view, text=text, pickable=True, size=[0.08, 0.04], pos=[0.08 * i, 0.06],
                        anchor=[0, 0]))
        overlay_chart_text_box_list[i].mouse['LeftButton']['Press'] = [LeftButtonPress]

    main = vmi.Main(return_globals)
    main.setAppName('pelvic_mark')
    main.setAppVersion(vmi.version)
    main.excludeKeys += ['main']

    menu_input = main.menuBar().addMenu('输入')
    menu_input.addAction('DICOM').triggered.connect(input_dicom)
    menu_input.addAction('Series').triggered.connect(input_series)

    main.layout().addWidget(voi_view, 0, 0, 3, 1)

    main.layout().addWidget(image_view, 0, 1, 3, 1)
    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import pathlib
import math
import tempfile
import time

import nibabel as nib
import numpy as np
import pydicom
import vtk
from PySide2.QtGui import *
from typing import Union, Optional, List
from numba import jit
from acetabular_segment import bulid_model, get_center, get_radius

import vmi


def ary_flip_pre(ary: np.ndarray):
    return np.flip(np.flip(ary.swapaxes(0, 2), 1), 0)


def pre_flip_ary(ary: np.ndarray):
    return np.flip(np.flip(ary, 0), 1).swapaxes(0, 2)


def bone_win_nor(image_data):
    image_data[image_data < -100] = -100
    image_data[image_data > 900] = 900
    image_data = (image_data + 100) / 1000
    return image_data


@jit(nopython=True)
def ary_pad(input_ary, constant):
    for k in [0, input_ary.shape[0] - 1]:
        for j in range(input_ary.shape[1]):
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in [0, input_ary.shape[1] - 1]:
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in range(input_ary.shape[1]):
            for i in [0, input_ary.shape[2] - 1]:
                input_ary[k][j][i] = constant


def imReslice(image: vtk.vtkImageData,
              cs: Optional[vmi.CS4x4] = None,
              size: Union[np.ndarray, List[float]] = None,
              back_scalar: Union[int, float] = 0) -> vtk.vtkImageData:
    if cs is None:
        cs = vmi.CS4x4()

    if size is None:
        size = [vmi.imSize_Vt(image, cs.axis(0)),
                vmi.imSize_Vt(image, cs.axis(1)),
                vmi.imSize_Vt(image, cs.axis(2))]

    spc = np.array([1, 1, 1])

    ext = [0, int(size[0]) - 1,
           0, int(size[1]) - 1,
           0, int(size[2]) - 1]

    re = vtk.vtkImageReslice()
    re.SetInputData(image)
    re.SetInterpolationModeToCubic()
    re.SetResliceAxes(cs.matrix4x4())
    re.SetOutputOrigin([0, 0, 0])
    re.SetOutputSpacing(spc)
    re.SetOutputExtent(ext)
    re.SetBackgroundLevel(back_scalar)
    re.Update()

    return re.GetOutput()


def modelToLargestRegion(pd: vtk.vtkPolyData):
    pdcf = vtk.vtkPolyDataConnectivityFilter()
    pdcf.SetInputData(pd)
    pdcf.SetExtractionModeToLargestRegion()
    pdcf.Update()

    smooth = vtk.vtkSmoothPolyDataFilter()
    smooth.SetInputData(pdcf.GetOutput())
    smooth.SetNumberOfIterations(20)
    smooth.Update()

    sinc = vtk.vtkWindowedSincPolyDataFilter()
    sinc.SetInputData(smooth.GetOutput())
    sinc.SetNumberOfIterations(20)
    sinc.SetBoundarySmoothing(0)
    sinc.SetFeatureEdgeSmoothing(0)
    sinc.SetFeatureAngle(60)
    sinc.SetPassBand(0.1)
    sinc.SetNonManifoldSmoothing(1)
    sinc.SetNormalizeCoordinates(1)
    sinc.Update()

    return sinc.GetOutput()


def read_dicom():
    dcmdir = vmi.askDirectory('DICOM')  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            global series
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                with pydicom.dcmread(series.filenames()[0]) as ds:
                    global patient_name
                    patient_name = str(ds.PatientName)
                    patient_name_box.draw_text('姓名：{}'.format(patient_name))
                    patient_name_box.setVisible(True)

                    global dicom_ds
                    dicom_ds = series.toKeywordValue()
                return series.read()  # 读取DICOM系列为图像数据


def on_init_voi():
    global voi_image
    voi_image = original_slice.data()
    voi_volume.setData(voi_image)
    voi_view.setCamera_Coronal()
    voi_view.setCamera_FitAll()


def init_pelvis():
    global model
    if td_pts[0].sum() != 0 and td_pts[1].sum() != 0:
        cs = vmi.CS4x4(axis0=np.array([1, 0, 0]),
                       axis1=np.array([0, 1, 0]),
                       axis2=np.array([0, 0, 1]),
                       origin=vmi.imOrigin(voi_image))

        seg_origin_pts = [td_pts[0] - 64 * cs.axis(0) - 48 * cs.axis(1) - 32 * cs.axis(2),
                          td_pts[1] - 32 * cs.axis(0) - 48 * cs.axis(1) - 32 * cs.axis(2)]

        seg_cs = [vmi.CS4x4(axis0=np.array([1, 0, 0]),
                            axis1=np.array([0, 1, 0]),
                            axis2=np.array([0, 0, 1]),
                            origin=seg_origin_pts[0]),
                  vmi.CS4x4(axis0=np.array([1, 0, 0]),
                            axis1=np.array([0, 1, 0]),
                            axis2=np.array([0, 0, 1]),
                            origin=seg_origin_pts[1])]

        size = np.array([96, 96, 96])
        seg_image = [imReslice(voi_image, seg_cs[0], size),
                     imReslice(voi_image, seg_cs[1], size)]

        seg_image[0].SetOrigin(seg_cs[0].origin())
        seg_image[1].SetOrigin(seg_cs[1].origin())

        model = bulid_model('nomal_model.h5')

        seg_ary = [vmi.imArray_VTK(seg_image[0]),
                   vmi.imArray_VTK(seg_image[1])]
        seg_pre = [np.squeeze(model.predict(bone_win_nor(ary_flip_pre(seg_ary[0]))[np.newaxis, :, :, :, np.newaxis])),
                   np.squeeze(model.predict(bone_win_nor(ary_flip_pre(seg_ary[1]))[np.newaxis, :, :, :, np.newaxis]))]
        seg_pre[0] = pre_flip_ary(seg_pre[0]) * (bone_value / probability)
        seg_pre[1] = pre_flip_ary(seg_pre[1]) * (bone_value / probability)

        voi_ext = np.array([vmi.imExtent(voi_image)[1] + 1,
                            vmi.imExtent(voi_image)[3] + 1,
                            vmi.imExtent(voi_image)[5] + 1])

        voi_ary = vmi.imArray_VTK(imReslice(voi_image, cs, voi_ext))

        seg_ijk_ori = [np.array((seg_origin_pts[0] - vmi.imOrigin(voi_image))).astype('int16'),
                       np.array((seg_origin_pts[1] - vmi.imOrigin(voi_image))).astype('int16')]

        voi_ary[seg_ijk_ori[0][2]:seg_ijk_ori[0][2] + 96,
        seg_ijk_ori[0][1]:seg_ijk_ori[0][1] + 96,
        seg_ijk_ori[0][0]:seg_ijk_ori[0][0] + 96] = seg_pre[0]

        voi_ary[seg_ijk_ori[1][2]:seg_ijk_ori[1][2] + 96,
        seg_ijk_ori[1][1]:seg_ijk_ori[1][1] + 96,
        seg_ijk_ori[1][0]:seg_ijk_ori[1][0] + 96] = seg_pre[1]

        pre_image = vmi.imVTK_Array(voi_ary, [0, 0, 0], [1, 1, 1])
        pre_image = vmi.imResample_Isotropic(pre_image)
        pd = vmi.imIsosurface(pre_image, bone_value)

        pelvis_prop.setData(modelToLargestRegion(pd))
        pelvis_view.setCamera_Coronal()
        pelvis_view.setCamera_FitAll()

        model.load_weights('acetabular_cup_placement_model.h5')
        pre = [np.squeeze(model.predict(bone_win_nor(ary_flip_pre(seg_ary[0]))[np.newaxis, :, :, :, np.newaxis])),
               np.squeeze(model.predict(bone_win_nor(ary_flip_pre(seg_ary[1]))[np.newaxis, :, :, :, np.newaxis]))]
        pre[0] = pre_flip_ary(pre[0])
        pre[1] = pre_flip_ary(pre[1])

        acetabular_center = [get_center(pre[0]), get_center(pre[1])]
        acetabular_radius = [get_radius(pre[0]), get_radius(pre[1])]

        acetabular_center[0] += seg_ijk_ori[0]
        acetabular_center[1] += seg_ijk_ori[1]

        cup_prop[0].setData(vmi.pdSphere(acetabular_radius[0], acetabular_center[0]))
        cup_prop[1].setData(vmi.pdSphere(acetabular_radius[1], acetabular_center[1]))

        # # 保存
        # nib.save(nib.Nifti1Image(bone_win_nor(ary_flip_pre(seg_ary[0])), None), 'p0.nii.gz')
        # nib.save(nib.Nifti1Image(bone_win_nor(ary_flip_pre(seg_ary[1])), None), 'p1.nii.gz')
        # nib.save(nib.Nifti1Image(seg_pre[0], None), '0.nii.gz')
        # nib.save(nib.Nifti1Image(seg_pre[1], None), '1.nii.gz')
        # nib.save(nib.Nifti1Image(pre[0], None), 's0.nii.gz')
        # nib.save(nib.Nifti1Image(pre[1], None), 's1.nii.gz')


def LeftButtonPress(**kwargs):
    global td_pts
    if kwargs['picked'] is voi_view:
        if tdpick_box.text() == '请选取右侧泪滴':
            td_pts[0] = voi_view.pickPt_Cell()
            td_prop[0].setData(vmi.pdSphere(2.5, td_pts[0]))
            tdpick_box.draw_text('请选取左侧泪滴')
        elif tdpick_box.text() == '请选取左侧泪滴':
            td_pts[1] = voi_view.pickPt_Cell()
            td_prop[1].setData(vmi.pdSphere(2.5, td_pts[1]))
            tdpick_box.draw_text('选取泪滴标志点')


def LeftButtonPressRelease(**kwargs):
    global bone_value
    if kwargs['picked'] is dicom_box:
        original_image = read_dicom()
        if original_image is not None:
            original_slice.setData(original_image)
            original_slice.setSlicePlaneOrigin_Center()
            original_slice.setSlicePlaneNormal_Coronal()
            original_view.setCamera_Coronal()
            original_view.setCamera_FitAll()
            on_init_voi()

        return True
    elif kwargs['picked'] is bone_value_box:
        v = vmi.askInt(-1000, bone_value, 3000)
        if v is not None:
            bone_value = v
            bone_value_box.draw_text('骨阈值：{:.0f} HU'.format(bone_value))
            voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})

        return True
    elif kwargs['picked'] is tdpick_box:
        if tdpick_box.text() == '选取泪滴标志点':
            tdpick_box.draw_text('请选取右侧泪滴')

        return True
    elif kwargs['picked'] is init_pelvis_box:
        init_pelvis_box.draw_text('请耐心等待')
        init_pelvis()
        init_pelvis_box.draw_text('创建骨盆')
        return True


def NoButtonWheel(**kwargs):
    global bone_value
    if kwargs['picked'] is bone_value_box:
        bone_value = min(max(bone_value + 10 * kwargs['delta'], -1000), 3000)
        bone_value_box.draw_text('骨阈值：{:.0f} HU'.format(bone_value))
        voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})


def return_globals():
    return globals()


if __name__ == '__main__':
    global original_view, voi_view, pelvis_view

    main = vmi.Main(return_globals)
    main.setAppName('全髋关节置换(THA)规划')
    main.setAppVersion(vmi.version)
    main.excludeKeys += ['main', 'i', 'box', 'prop', 'voi_image']

    select = vmi.askButtons(['新建', '打开'], title=main.appName())
    if select is None:
        vmi.appexit()

    if select == '打开':
        open_file = vmi.askOpenFile(nameFilter='*.vmi', title='打开存档')
        if open_file is not None:
            main.loads(pathlib.Path(open_file).read_bytes())
        else:
            vmi.appexit()
    elif select == '新建':
        patient_name = str()

        # 0 原始图像
        original_target, LR = np.zeros(3), 1
        original_view = vmi.View()

        original_slice = vmi.ImageSlice(original_view)  # 断层显示
        original_slice.setColorWindow_Bone()
        original_slice.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        dicom_box = vmi.TextBox(original_view, text='DICOM', pickable=True,
                                size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0])
        dicom_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        patient_name_box = vmi.TextBox(original_view, text='姓名', visible=False,
                                       size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0])

        # 1 目标区域
        voi_view = vmi.View()
        voi_view.mouse['LeftButton']['Press'] = [LeftButtonPress]

        bone_value, target_value = 200, -1100

        bone_value_box = vmi.TextBox(voi_view, text='骨阈值：{:.0f} HU'.format(bone_value),
                                     size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0], pickable=True)
        bone_value_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        bone_value_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        voi_size, voi_center, voi_origin, voi_cs = np.zeros(3), np.zeros(3), np.zeros(3), vmi.CS4x4()
        voi_image = vtk.vtkImageData()

        voi_volume = vmi.ImageVolume(voi_view, pickable=True)
        voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})
        voi_volume.setColor({bone_value: [1, 1, 0.6]})

        # 泪滴选点
        td_pts = np.zeros((2, 3))
        td_prop = [vmi.PolyActor(voi_view, color=[0.4, 0.6, 1]),
                   vmi.PolyActor(voi_view, color=[0.4, 0.6, 1])]

        tdpick_box = vmi.TextBox(voi_view, text='选取泪滴标志点', size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0], pickable=True)
        tdpick_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        probability = 0.6

        # 创建骨盆
        init_pelvis_box = vmi.TextBox(voi_view, text='创建骨盆', fore_color=QColor('white'), back_color=QColor('crimson'),
                                      bold=True, size=[0.2, 0.04], pos=[1, 0.04], anchor=[1, 0], pickable=True)
        init_pelvis_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        # 5 骨盆视图
        pelvis_view = vmi.View()
        pelvis_prop = vmi.PolyActor(pelvis_view, color=[1, 1, 0.6])

        cup_prop = [vmi.PolyActor(pelvis_view, color=[0.4, 0.6, 1], line_width=3, pickable=True),
                    vmi.PolyActor(pelvis_view, color=[0.4, 0.6, 1], line_width=3, pickable=True)]

    # 视图布局
    for v in [original_view, voi_view, pelvis_view]:
        v.setMinimumWidth(round(0.5 * main.screenWidth()))

    original_view.setParent(main.scrollArea())
    main.layout().addWidget(original_view, 0, 0, 1, 1)
    main.layout().addWidget(voi_view, 0, 1, 1, 1)
    main.layout().addWidget(pelvis_view, 0, 2, 1, 1)

    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import datetime
import os
import pathlib
import shutil
import tempfile
import threading
from typing import List, Optional, Dict

import SimpleITK as sitk
import numpy as np
import pydicom
import vtk
from PySide2.QtCore import *

import vmi


class SeriesData(QObject, vmi.Menu):
    """DICOM系列数据"""

    def __init__(self, filenames=None, name=None):
        QObject.__init__(self)

        self.name = name if name else self.tr('系列 (Series)')
        vmi.Menu.__init__(self)

        self._Filenames = filenames
        self._Directory = tempfile.TemporaryDirectory()

        if self._Filenames and len(self._Filenames) > 0:
            self._Filenames = [shutil.copy2(f, self._Directory.name) for f in self._Filenames]

    def __setstate__(self, s):
        self.__init__(name=s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Filenames = []
        self._Directory = tempfile.TemporaryDirectory()
        for kw in s['_Bytes']['_Files']:
            p = pathlib.Path(self._Directory.name) / kw
            p.write_bytes(s['_Bytes']['_Files'][kw])
            self._Filenames.append(str(p))

        del s['_Bytes']

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['_Filenames', '_Directory', 'menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Bytes'] = {'_Files': {pathlib.Path(f).name: pathlib.Path(f).read_bytes() for f in self._Filenames}}
        return s

    def __str__(self):
        with pydicom.dcmread(self._Filenames[0]) as ds:
            r = ['', '', '', '', '']
            for e in [_ for _ in ds if _.VR != 'SQ']:
                e.showVR = False
                if e.name in ['Modality', 'Slice Thickness', 'Rows', 'Columns', 'Pixel Spacing', 'Pixel Data']:
                    r[0] += repr(e) + '\n'
                elif e.name.startswith('Patient'):
                    r[1] += repr(e) + '\n'
                elif e.name.startswith('Study'):
                    r[2] += repr(e) + '\n'
                elif e.name.startswith('Series'):
                    r[3] += repr(e) + '\n'
                elif e.name.startswith('Image'):
                    r[4] += repr(e) + '\n'
            return 'Slices: {}\n'.format(len(self._Filenames)) + r[0] + r[1] + r[2] + r[3] + r[4]

    def __repr__(self):
        return self.__str__()

    def toKeywordValue(self) -> Dict[str, str]:
        d = {'Slices': len(self._Filenames)}
        with pydicom.dcmread(self._Filenames[0]) as ds:
            for e in ds.iterall():
                if e.value.__class__ in [None, bool, int, float, str]:
                    d[e.keyword] = e.value
                elif e.value.__class__ in [bytes]:
                    d[e.keyword] = '{} bytes'.format(len(e.value))
                else:
                    d[e.keyword] = str(e.value)
        d = {kw: d[kw] for kw in sorted(d)}
        return d

    def filenames(self) -> List[str]:
        """返回文件列表"""
        return self._Filenames

    def readITK(self) -> sitk.Image:
        """读取系列为SimpleITK图像"""
        r = sitk.ImageSeriesReader()
        r.SetFileNames(self._Filenames)
        image = r.Execute()
        return image

    def read(self) -> Optional[vtk.vtkImageData]:
        """读取系列为vtkImageData图像"""
        image = None

        def func():
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.nii'

                w = sitk.ImageFileWriter()
                w.SetFileName(str(p))
                w.SetImageIO('NiftiImageIO')
                w.Execute(self.readITK())

                nonlocal image
                image = vmi.imOpenFile_NIFTI(str(p))

        t = threading.Thread(target=func)
        t.start()
        vmi.appwait(t)
        return image


def sortSeries(dcmdir: str, recursive: bool = True) -> List[SeriesData]:
    """
    识别并排序DICOM文件，返回SeriesData对象的列表

    :param dcmdir: DICOM文件夹
    :param recursive: 递归搜索
    :return: [SeriesData]
    """
    dcmdir = str(pathlib.Path(dcmdir))
    series = []

    def func():
        roots = [dcmdir]
        if recursive:
            roots = [root for root, *_ in os.walk(dcmdir)]
            roots.sort()
        for root in roots:
            with tempfile.TemporaryDirectory() as p:
                if vmi.contains_zh_CN(root):
                    for f in os.listdir(root):
                        f = os.path.join(root, f)
                        if os.path.isfile(f):
                            shutil.copy2(f, p)
                    root = p

                for i in sitk.ImageSeriesReader.GetGDCMSeriesIDs(root):
                    filenames = sitk.ImageSeriesReader.GetGDCMSeriesFileNames(root, i)
                    series.append(SeriesData(filenames))

    t = threading.Thread(target=func)
    t.start()
    vmi.appwait(t)
    return series


def sortFilenames(dcmdir: str, recursive: bool = True) -> List[List[str]]:
    """
    识别并排序DICOM文件，返回Series的文件名列表的列表

    :param dcmdir: DICOM文件夹
    :param recursive: 递归搜索
    :return: [[filenames]]
    """
    dcmdir = str(pathlib.Path(dcmdir))
    filenames_list = []

    roots = [dcmdir]
    if recursive:
        roots = [root for root, *_ in os.walk(dcmdir)]
        roots.sort()
    for root in roots:
        with tempfile.TemporaryDirectory() as p:
            if vmi.contains_zh_CN(root):
                for f in os.listdir(root):
                    f = os.path.join(root, f)
                    if os.path.isfile(f):
                        shutil.copy2(f, p)
                directory = p
            else:
                directory = root

            for i in sitk.ImageSeriesReader.GetGDCMSeriesIDs(directory):
                filenames = sitk.ImageSeriesReader.GetGDCMSeriesFileNames(directory, i)
                filenames = [str(pathlib.Path(root) / pathlib.Path(f).name) for f in filenames]
                filenames_list.append(filenames)
                print('sortFilenames {}'.format(len(filenames_list)))
    return filenames_list


def imSaveDICOM_SeriesCT(image: vtk.vtkImageData, path=None) -> None:
    """保存vtkImageData到DICOM文件"""
    if path is None:
        path = vmi.askDirectory('导出 (Export) DICOM Series')
        if path is None:
            return

    date = datetime.datetime.now().strftime('%Y%m%d')
    time = datetime.datetime.now().strftime('%H%M%S.%f')
    pixel_array = vmi.imArray_VTK(image)
    origin = np.array(image.GetOrigin())
    dimensions = np.array(image.GetDimensions())

    for k in pixel_array.shape[0]:
        suffix = '.dcm'
        filename_little_endian = tempfile.NamedTemporaryFile(suffix=suffix).name
        filename_big_endian = tempfile.NamedTemporaryFile(suffix=suffix).name

        file_meta = pydicom.Dataset()
        file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
        file_meta.MediaStorageSOPInstanceUID = "1.2.3"
        file_meta.ImplementationClassUID = "1.2.3.4"

        ds = pydicom.FileDataset(filename_little_endian, {}, file_meta=file_meta, preamble=b"\0" * 128)
        ds.is_little_endian = True
        ds.is_implicit_VR = True

        ds.PatientName = 'anonymous'
        ds.PatientID = '000000'

        ds.AccessionNumber = '0000000000000'
        ds.AcquisitionDate = date
        ds.AcquisitionDateTime = date + time
        ds.AcquisitionNumber = ''
        ds.AcquisitionTime = time

        ds.BitsAllocated = 16
        ds.BitsStored = 12
        ds.HighBit = 11

        ds.ContentDate = date
        ds.ContentTime = time

        ds.ImageOrientationPatient = ['1', '0', '0', '0', '1', '0']
        ds.ImageType = ['ORIGINAL', 'PRIMARY', 'AXIAL', 'HELIX']

        ds.Rows = dimensions[0]
        ds.Columns = dimensions[1]
        print(pixel_array.dtype)
        ds.PixelData = pixel_array.tobytes()

        ds.InstanceCreationDate = date
        ds.InstanceCreationTime = time


if __name__ == '__main__':
    dcmdir = vmi.askDirectory('DICOM')

    if dcmdir is not None:
        series_list = vmi.sortSeries(dcmdir)
        series = vmi.askSeries(series_list)
        if series is not None:
            image = series.read()
            print(image.GetOrigin())

    imSaveDICOM_SeriesCT(image)
['-233.5', '-92.5', '386.099976']
(233.5, 92.5, 386.0999755859375)



import csv
import hashlib
import time
from typing import Optional, Union

import gridfs
import pydicom
import pymongo
import requests
from PySide2.QtWidgets import *

import vmi

client = pymongo.MongoClient('mongodb://skyward:Yw309309@med-3d.top:27018/skyward', 27017)
database = client.skyward
case_db = database.get_collection('case')
order_db = database.get_collection('order')
user_db = database.get_collection('user')
role_customer_db = database.get_collection('role_customer')
role_sales_worker_db = database.get_collection('role_sales_worker')
role_product_designer_db = database.get_collection('role_product_designer')
role_quality_inspector_db = database.get_collection('role_quality_inspector')
dicom_fs = gridfs.GridFS(database, collection='dicom')
vti_fs = gridfs.GridFS(database, collection='vti')
vtp_fs = gridfs.GridFS(database, collection='vtp')
stl_fs = gridfs.GridFS(database, collection='stl')


def ask_user() -> Optional[dict]:
    ok, find_user = True, {}
    while ok:
        text, ok = QInputDialog.getText(None, '请输入', '手机号：', QLineEdit.Normal)
        if not ok:
            return None

        find_user = user_db.find_one({'telephone': text})
        if not find_user:
            vmi.askInfo('手机号未注册')
            ok = True
            continue

        ok = False

    ok = True
    while ok:
        text, ok = QInputDialog.getText(None, '请输入', '密码：', QLineEdit.Password)
        if not ok:
            return None

        if find_user['password'] != hashlib.md5(text.encode()).hexdigest():
            vmi.askInfo('密码错误')
            ok = True
            continue

        ok = False
    return find_user


def ask_customer_id() -> Optional[dict]:
    sizep = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

    dialog = QDialog()
    dialog.setSizePolicy(sizep)
    dialog.setWindowTitle('选择顾客')
    dialog.setLayout(QVBoxLayout())
    dialog.setMinimumWidth(200)

    customer_combo = QComboBox()
    dialog.layout().addWidget(customer_combo)

    for i in role_customer_db.find():
        customer_combo.addItem('{} {} {}'.format(i['_id'], i['name'], i['organization']), i['_id'])

    ok_cancel = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
    ok_cancel.setSizePolicy(sizep)
    ok_cancel.accepted.connect(dialog.accept)
    ok_cancel.rejected.connect(dialog.reject)
    dialog.layout().addWidget(ok_cancel)

    if dialog.exec_() == QDialog.Accepted:
        return customer_combo.currentData()
    else:
        return None


def ask_sales_worker_id() -> Optional[dict]:
    sizep = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

    dialog = QDialog()
    dialog.setSizePolicy(sizep)
    dialog.setWindowTitle('选择销售')
    dialog.setLayout(QVBoxLayout())
    dialog.setMinimumWidth(200)

    sales_worker_combo = QComboBox()
    dialog.layout().addWidget(sales_worker_combo)

    for i in role_sales_worker_db.find():
        sales_worker_combo.addItem('{} {} {}'.format(i['_id'], i['name'], i['organization']), i['_id'])

    ok_cancel = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
    ok_cancel.setSizePolicy(sizep)
    ok_cancel.accepted.connect(dialog.accept)
    ok_cancel.rejected.connect(dialog.reject)
    dialog.layout().addWidget(ok_cancel)

    if dialog.exec_() == QDialog.Accepted:
        return sales_worker_combo.currentData()
    else:
        return None


def dicom_kw_value(f: str) -> dict:
    kw_value = {}
    with pydicom.dcmread(f) as ds:
        for e in ds.iterall():
            if e.value.__class__ in [None, bool, int, float, str]:
                kw_value[e.keyword] = e.value
            elif e.value.__class__ in [bytes]:
                kw_value[e.keyword] = '{} bytes'.format(len(e.value))
            else:
                kw_value[e.keyword] = str(e.value)
    return {kw: kw_value[kw] for kw in sorted(kw_value)}


def case(order_model: str, order_uid: str, order_files: list,
         patient_name: str, patient_sex: str, patient_age: str,
         dicom_files: dict, vti_files: dict, vtp_files: dict, stl_files: dict,
         **kwargs) -> Optional[dict]:
    """
    构造本地案例
    :param order_model: 产品型号
    :param order_uid: 产品序列号
    :param order_files: 产品文件列表
    :param patient_name: 患者姓名
    :param patient_sex: 患者性别
    :param patient_age: 患者年龄
    :param dicom_files: 原始DICOM文件，每个关键词索引一个文件列表
    :param vti_files: vtkXMLImageData格式文件，每个关键词索引一个文件
    :param vtp_files: vtkXMLPolyData格式文件，每个关键词索引一个文件
    :param stl_files: STL格式文件，每个关键词索引一个文件
    :param kwargs: 其它相关信息，每个关键词索引一个支持数据库直接录入的对象
    :return: 案例信息
    """
    case_data = {}

    # 上传原始DICOM文件
    for kw in dicom_files:
        sop_uids = []
        for f in dicom_files[kw]:
            sop_uid = pydicom.dcmread(f).SOPInstanceUID
            sop_uids.append(sop_uid)

            if not dicom_fs.exists({'SOPInstanceUID': sop_uid}):
                dicom_fs.put(open(f, 'rb').read(), **dicom_kw_value(f))
        case_data[kw] = sop_uids

    # 上传其它格式文件
    files = [vti_files, vtp_files, stl_files]
    fs = [vti_fs, vtp_fs, stl_fs]
    for i in range(len(files)):
        for kw in files[i]:
            f_bytes = open(files[i][kw], 'rb').read()
            md5 = hashlib.md5(f_bytes).hexdigest()

            if fs[i].exists({'md5': md5}):
                case_data[kw] = fs[i].find_one({'md5': md5})._id
            else:
                case_data[kw] = fs[i].put(f_bytes)

    # 上传其它信息
    case_data['order_model'] = order_model
    case_data['order_uid'] = order_uid
    case_data['order_files'] = order_files
    case_data['patient_name'] = patient_name
    case_data['patient_sex'] = patient_sex
    case_data['patient_age'] = patient_age

    for kw in kwargs:
        case_data[kw] = kwargs[kw]
    return case_data


def case_sync(case: dict, ukw: list) -> Union[dict, bool]:
    """
    同步案例
    :param case: 本地案例信息
    :param ukw: 特征键列表，用于判断相同案例
    :return: 同步后的案例信息
    """
    ukw_value = {kw: case[kw] for kw in ukw}

    # 清理相同无效案例
    for c in case_db.find(ukw_value):
        if 'case_uid' not in c:
            case_db.delete_one({'_id': c['_id']})

    # 查找相同有效案例
    find = case_db.find_one(ukw_value)

    if find and 'case_uid' in find:
        if find['order_released']:
            vmi.askInfo('生产单已发布')
            r = vmi.askButtons(['下载案例'])
        elif find['customer_confirmed']:
            vmi.askInfo('顾客已确认')
            r = vmi.askButtons(['上传案例', '下载案例', '发布生产单'])
        else:
            r = vmi.askButtons(['上传案例', '下载案例'])
    else:
        r = vmi.askButtons(['创建案例'])

    if r == '创建案例':
        user = ask_user()
        if not user:
            return False

        role_product_designer_id = user['role_product_designer_id']
        print(role_product_designer_id)
        if not role_product_designer_db.find_one({'_id': role_product_designer_id}):
            vmi.askInfo('错误：只有产品设计师可以创建案例')
            return False

        customer_uid = ask_customer_id()
        if not customer_uid:
            return False

        sales_worker_uid = ask_sales_worker_id()
        if not sales_worker_uid:
            return False

        case['role_product_designer_id'] = role_product_designer_id
        case['role_customer_id'] = customer_uid
        case['role_sales_worker_id'] = sales_worker_uid
        case['role_quality_inspector_id'] = None
        case['customer_confirmed'] = False
        case['order_released'] = False
        case['order_inspected'] = False
        case['order_completed'] = False

        _id = case_db.insert_one({kw: case[kw] for kw in sorted(case)}).inserted_id
        try:
            case_uid = vmi.convert_base(round(time.time() * 100))
            response = requests.post(url='http://med-3d.top:2334/sys/caseCreate',
                                     headers={'_id': str(_id), 'case_uid': case_uid})
            if response.text == 'true':
                find = case_db.find_one({'_id': _id})
                if 'case_uid' in find:
                    return True
                else:
                    case_db.delete_one({'_id': _id})
                    vmi.askInfo('错误：云端未分配编号')
            else:
                case_db.delete_one({'_id': _id})
                vmi.askInfo('错误：' + response.text)
        except Exception as e:
            case_db.delete_one({'_id': _id})
            vmi.askInfo('错误：' + str(e))
    elif r == '上传案例':
        user = ask_user()
        if not user:
            return False

        role_product_designer_id = user['role_product_designer_id']
        if not role_product_designer_db.find_one({'_id': role_product_designer_id}):
            vmi.askInfo('错误：只有产品设计师可以上传案例')
            return False

        case['role_product_designer_id'] = role_product_designer_id
        case['customer_confirmed'] = False
        case['order_released'] = False
        case['order_inspected'] = False
        case['order_completed'] = False

        try:
            case_uid = vmi.convert_base(round(time.time() * 100))
            response = requests.post(url='http://med-3d.top:2334/sys/caseUpdate',
                                     headers={'_id': str(find['_id']), 'case_uid': case_uid})
            if response.text == 'true':
                case_db.find_one_and_update({'_id': find['_id']}, {'$set': case})
                return True
            else:
                vmi.askInfo('错误：' + response.text)
        except Exception as e:
            vmi.askInfo('错误：' + str(e))
    elif r == '下载案例':
        return case_db.find_one({'_id': find['_id']})
    elif r == '发布生产单':
        user = ask_user()
        if not user:
            return False

        role_product_designer_id = user['role_product_designer_id']
        if not role_product_designer_db.find_one({'_id': role_product_designer_id}):
            vmi.askInfo('错误：只有产品设计师可以发布生产单')
            return False

        case_set = {'role_product_designer_id': role_product_designer_id, 'order_released': True}

        try:
            case_uid = vmi.convert_base(round(time.time() * 100))
            response = requests.post(url='http://med-3d.top:2334/sys/caseUpdate',
                                     headers={'_id': str(find['_id']), 'case_uid': case_uid})
            if response.text == 'true':
                case_db.find_one_and_update({'_id': find['_id']}, {'$set': case_set})
                return True
            else:
                vmi.askInfo('错误：' + response.text)
        except Exception as e:
            vmi.askInfo('错误：' + str(e))


def csv_update(db: str):
    f = vmi.askOpenFile('*.csv', db)
    if f is None:
        return

    db = database.get_collection(db)

    with open(f, newline='') as f:
        kw = []
        for row in csv.reader(f):
            if len(kw):
                kw[0] = '_id'
                i = dict(zip(kw, row))
                db.find_one_and_update({'_id': i['_id']}, {'$set': i}, upsert=True)
            else:
                kw = row.copy()


if __name__ == '__main__':
    csv_update('role_customer')
    csv_update('role_sales_worker')
    csv_update('role_product_designer')
    csv_update('role_quality_inspector')
